"""
Copyright (c) 2023 Gabriel Guerrer

Distributed under the MIT license - See LICENSE for details
"""

from .rava_cfg import RAVA_CFG
from .rava_subapp import RAVA_SUBAPP
from .rava_app import RAVA_APP

from . import acq
from . import ctrlp


"""
Copyright (c) 2023 Gabriel Guerrer

Distributed under the MIT license - See LICENSE for details
"""

"""
The RAVA_RNG_AIO class offers the same functionality as RAVA_RNG but within an
asynchronous framework.
"""

import asyncio
import numpy as np

from rng_rava import RAVA_RNG
from rng_rava import find_rava_port, version_a_greater_b, print_health_startup_results, bytes_to_list, bytes_to_array
from rng_rava.rava_defs import *


### RAVA_RNG_AIO

class RAVA_RNG_AIO(RAVA_RNG):

    def __init__(self):
        super().__init__(dev_name='RAVA_AIO')
        self.queue_type = asyncio.Queue

        # Variables
        self.serial_listen_task = None


    async def connect(self, serial_number):
        # Debug
        self.lg.debug('> {} CONNECT'.format(self.dev_name))

        # Find serial port
        port = find_rava_port(serial_number)
        if port is None:
            self.lg.error('{} Connect: No device found with SN {}'.format(self.dev_name, serial_number))
            return False

        # Open serial connection
        if not self.open_serial(port):
            return False

        # Reset serial data queues
        self.init_queue_data()

        # Start listening for serial commands
        self.serial_listen_task = asyncio.create_task(self.loop_serial_listen())

        # Stop any active RNG byte stream
        self.snd_rng_byte_stream_stop()

        # Request firmware info
        self.firmware_dict = await self.get_eeprom_firmware()
        self.firmware_version =  self.firmware_dict['version']
        self.health_startup_enabled = self.firmware_dict['health_startup_enabled']
        self.health_continuous_enabled = self.firmware_dict['health_continuous_enabled']
        self.led_enabled = self.firmware_dict['led_enabled']
        self.lamp_enabled = self.firmware_dict['lamp_enabled']
        self.peripherals_enabled = self.firmware_dict['peripherals_enabled']

        # Check Firmware version
        if version_a_greater_b(FIRMWARE_MIN_VERSION, self.firmware_version):
            self.lg.error('{} Connect: Firmware version v{} is incompatible with the driver'.format(self.dev_name, self.firmware_version))
            return False

        # Print connection info
        self.lg.info('{} Connect: Success'
                '\n{}  {}, Firmware v{}, SN={}, at {}'
                .format(self.dev_name, LOG_FILL,
                        self.dev_usb_name, self.firmware_version, self.dev_serial_number, self.serial.port))

        # Request Health startup info
        if self.health_startup_enabled:
            self.health_startup_success, test_vars = await self.get_health_startup_results()

            # Print test info
            print_health_startup_results(self.health_startup_success, test_vars)

            # Error? Users have then a limited command variety (see Firmware code)
            if not self.health_startup_success:
                self.lg.error('{} Connect: Startup tests failed'.format(self.dev_name))
                return False

        # Reset Health continuous erros
        self.get_health_continuous_errors()

        return True


    ## QUEUE

    def put_queue_data(self, comm, value, comm_ext_id=0):
        # Check key
        if comm not in self.serial_data:
            self.lg.error('{} Data: Unknown comm {}'.format(self.dev_name, comm))
            return False

        # Extended key
        if comm_ext_id:
            comm_ext = '{}_{}'.format(comm, comm_ext_id)

            # Queue exists?
            if comm_ext not in self.serial_data:
                self.serial_data[comm_ext] = self.queue_type()
        else:
            comm_ext = comm

        # Try writing to queue
        try:
            self.serial_data[comm_ext].put_nowait(value)
            return True

        except asyncio.QueueFull:
            self.lg.error('{} Data: {} Queue full'.format(self.dev_name, comm_ext))
            return False


    async def get_queue_data(self, comm, comm_ext_id=0):
        # Check key
        if comm not in self.serial_data:
            self.lg.error('{} Data: Unknown comm {}'.format(self.dev_name, comm))
            return None

        # Extended key
        if comm_ext_id:
            comm_ext = '{}_{}'.format(comm, comm_ext_id)

            # Queue exists?
            if comm_ext not in self.serial_data:
                self.serial_data[comm_ext] = self.queue_type()
        else:
            comm_ext = comm

        # Read asyncio queue
        return await self.serial_data[comm_ext].get()


    ## SERIAL

    async def loop_serial_listen(self):
        # Debug
        self.lg.debug('> {} LOOP SERIAL LISTEN'.format(self.dev_name))

        # Loop while connected
        while self.serial_connected.is_set():

            # Command available?
            comm_inwaiting = self.inwaiting_serial()
            if comm_inwaiting is None:
                continue # Disconnected

            if comm_inwaiting > 0:

                # Read command starting char
                comm_start = self.read_serial(1)
                if comm_start is None:
                    continue # Disconnected

                # Starts with $?
                if comm_start == COMM_MSG_START:

                    # Read remaining command bytes
                    comm_msg = self.read_serial(COMM_MSG_LEN-1)
                    if comm_msg is None:
                        continue # Disconnected

                    comm_id = comm_msg[0]
                    comm_data = comm_msg[1:]

                    # Known command id?
                    if comm_id in D_DEV_COMM_INV:

                        # Debug
                        self.lg.debug('> COMM RCV {}'.format([D_DEV_COMM_INV[comm_id], *[c for c in comm_data]]))

                        # Process Command
                        try:
                            self.process_serial_comm(comm_id, comm_data)

                        except Exception as err:
                            self.lg.error('{} Serial Listen Loop: Error processing command_id {}'
                                        '\n{}  {} - {}'
                                        .format(self.dev_name, comm_id, LOG_FILL, type(err).__name__, err))

                            # Close device
                            self.close()

                    else:
                        self.lg.warning('{} Serial Listen Loop: Unknown command_id {}'.format(self.dev_name, comm_id))

                else:
                    self.lg.warning('{} Serial Listen Loop: Commands must start with {}'.format(self.dev_name, COMM_MSG_START))

            # The non-blocking method is prefered for finishing the thread
            # when closing the device
            else:
                await asyncio.sleep(SERIAL_LISTEN_LOOP_INTERVAL_S)


    ## DEVICE

    async def get_device_serial_number(self):
        comm = 'DEVICE_SERIAL_NUMBER'
        if self.snd_rava_msg(comm):
            return await self.get_queue_data(comm)


    async def get_device_temperature(self):
        comm = 'DEVICE_TEMPERATURE'
        if self.snd_rava_msg(comm):
            return await self.get_queue_data(comm)


        comm = 'DEVICE_FREE_RAM'
        if self.snd_rava_msg(comm):
            return await self.get_queue_data(comm)


    ## EEPROM

    async def get_eeprom_device(self):
        comm = 'EEPROM_DEVICE'
        rava_send = True
        if self.snd_rava_msg(comm, [rava_send], 'B'):
            data_vars = await self.get_queue_data(comm)
            data_names = ['temp_calib_slope', 'temp_calib_intercept']
            return dict(zip(data_names, data_vars))


    async def get_eeprom_firmware(self):
        comm = 'EEPROM_FIRMWARE'
        if self.snd_rava_msg(comm):
            version_major, version_minor, version_patch, modules = await self.get_queue_data(comm)

            version = '{}.{}.{}'.format(version_major, version_minor, version_patch)
            health_startup_enabled = modules & 1 << 0 != 0
            health_continuous_enabled = modules & 1 << 1 != 0
            led_enabled = modules & 1 << 2 != 0
            lamp_enabled = modules & 1 << 3 != 0
            peripherals_enabled = modules & 1 << 4 != 0
            serial1_enabled = modules & 1 << 5 != 0
            return {'version':version,
                    'health_startup_enabled':health_startup_enabled,
                    'health_continuous_enabled':health_continuous_enabled,
                    'led_enabled':led_enabled,
                    'lamp_enabled':lamp_enabled,
                    'peripherals_enabled':peripherals_enabled,
                    'serial1_enabled':serial1_enabled
                    }


    async def get_eeprom_pwm(self):
        comm = 'EEPROM_PWM'
        rava_send = True
        if self.snd_rava_msg(comm, [rava_send], 'B'):
            freq_id, duty = await self.get_queue_data(comm)
            return {'freq_id':freq_id, 'freq_str':D_PWM_FREQ_INV[freq_id], 'duty':duty}


    async def get_eeprom_rng(self):
        comm = 'EEPROM_RNG'
        rava_send = True
        if self.snd_rava_msg(comm, [rava_send], 'B'):
            sampling_interval_us = await self.get_queue_data(comm)
            return {'sampling_interval_us':sampling_interval_us}


    async def get_eeprom_led(self):
        comm = 'EEPROM_LED'
        rava_send = True
        if self.snd_rava_msg(comm, [rava_send], 'B'):
            led_attached = await self.get_queue_data(comm)
            return {'led_attached':led_attached}


    async def get_eeprom_lamp(self):
        comm = 'EEPROM_LAMP'
        rava_send = True
        if self.snd_rava_msg(comm, [rava_send], 'B'):
            data_vars = await self.get_queue_data(comm)
            data_names = ['exp_dur_max_ms', 'exp_z_significant', 'exp_mag_smooth_n_trials']
            return dict(zip(data_names, data_vars))


    ## PWM

    async def get_pwm_setup(self):
        rava_send = True
        comm = 'PWM_SETUP'
        if self.snd_rava_msg(comm, [rava_send], 'B'):
            freq_id, duty = await self.get_queue_data(comm)
            return {'freq_id':freq_id, 'freq_str':D_PWM_FREQ_INV[freq_id], 'duty':duty}


    ## RNG

    async def get_rng_setup(self):
        comm = 'RNG_SETUP'
        rava_send = True
        if self.snd_rava_msg(comm, [rava_send], 'B'):
            sampling_interval_us = await self.get_queue_data(comm)
            return {'sampling_interval_us':sampling_interval_us}


    async def get_rng_pulse_counts(self, n_counts, output_type='array'):
        if n_counts == 0:
            self.lg.error('{} RNG PC: Provide n_counts > 0'.format(self.dev_name))
            return None, None
        if n_counts >= 2**16:
            self.lg.error('{} RNG PC: Provide n_counts as a 16-bit integer'.format(self.dev_name))
            return None, None

        comm = 'RNG_PULSE_COUNTS'
        if self.snd_rava_msg(comm, [n_counts], 'H'):
            counts_bytes_a, counts_bytes_b = await self.get_queue_data(comm)

            if (counts_bytes_a is None) or (counts_bytes_a is None):
                return None, None

            else:
                if output_type == 'list':
                    counts_a = bytes_to_list(counts_bytes_a, 'B')
                    counts_b = bytes_to_list(counts_bytes_b, 'B')
                    return counts_a, counts_b

                elif output_type == 'array':
                    counts_a = bytes_to_array(counts_bytes_a, np.uint8)
                    counts_b = bytes_to_array(counts_bytes_b, np.uint8)
                    return counts_a, counts_b

                else:
                    return counts_bytes_a, counts_bytes_b


    async def get_rng_bits(self, bit_source_id):
        if bit_source_id not in D_RNG_BIT_SRC_INV:
            self.lg.error('{} RNG Bits: Unknown bit_source_id {}'.format(self.dev_name, bit_source_id))
            return None

        comm = 'RNG_BITS'
        if self.snd_rava_msg(comm, [bit_source_id], 'B'):
            bit_source_id_recv, *bits = await self.get_queue_data(comm)

            if bit_source_id == D_RNG_BIT_SRC['AB_RND']:
                bit_type_str = D_RNG_BIT_SRC_INV[bit_source_id_recv]
                return [bit_type_str, bits[0]]
            elif bit_source_id == D_RNG_BIT_SRC['AB']:
                return list(bits)
            else:
                return bits[0]


    async def get_rng_bytes(self, n_bytes, postproc_id=D_RNG_POSTPROC['NONE'], request_id=0, output_type='array'):
        if n_bytes == 0:
            self.lg.error('{} RNG Bytes: Provide n_bytes > 0'.format(self.dev_name))
            return None, None
        if n_bytes >= 2**16:
            self.lg.error('{} RNG Bytes: Provide n_bytes as a 16-bit integer'.format(self.dev_name))
            return None, None
        if postproc_id not in D_RNG_POSTPROC_INV:
            self.lg.error('{} RNG Bytes: Unknown postproc_id {}'.format(self.dev_name, postproc_id))
            return None, None

        comm = 'RNG_BYTES'
        if self.snd_rava_msg(comm, [n_bytes, postproc_id, request_id], 'HBB'):
            rng_bytes_a, rng_bytes_b = await self.get_queue_data(comm, comm_ext_id=request_id)

            if (rng_bytes_a is None) or (rng_bytes_b is None):
                return None, None

            else:
                if output_type == 'list':
                    rng_a = bytes_to_list(rng_bytes_a, 'B')
                    rng_b = bytes_to_list(rng_bytes_b, 'B')
                    return rng_a, rng_b

                elif output_type == 'array':
                    rng_a = bytes_to_array(rng_bytes_a, np.uint8)
                    rng_b = bytes_to_array(rng_bytes_b, np.uint8)
                    return rng_a, rng_b

                else:
                    return rng_bytes_a, rng_bytes_b


    async def get_rng_int8s(self, n_ints, int_delta, output_type='array'):
        if n_ints == 0:
            self.lg.error('{} RNG Ints: Provide n_ints > 0'.format(self.dev_name))
            return None
        if n_ints >= 2**16:
            self.lg.error('{} RNG Ints: Provide n_ints as a 16-bit integer'.format(self.dev_name))
            return None
        if int_delta >= 2**8:
            self.lg.error('{} RNG Ints: Provide int_delta as a 8-bit integer'.format(self.dev_name))
            return None
        if int_delta == 0:
            self.lg.error('{} RNG Ints: Provide int_delta > 0'.format(self.dev_name))
            return None

        comm = 'RNG_INT8S'
        if self.snd_rava_msg(comm, [n_ints, int_delta], 'HB'):
            ints_bytes = await self.get_queue_data(comm)
            if ints_bytes is None:
                return None
            else:
                if output_type == 'list':
                    return bytes_to_list(ints_bytes, 'B')
                else:
                    return bytes_to_array(ints_bytes, np.uint8)


    async def get_rng_int16s(self, n_ints, int_delta, output_type='array'):
        # int_delta = int_max - int_min
        if n_ints == 0:
            self.lg.error('{} RNG Ints: Provide n_ints > 0'.format(self.dev_name))
            return None
        if n_ints >= 2**16:
            self.lg.error('{} RNG Ints: Provide n_ints as a 16-bit integer'.format(self.dev_name))
            return None
        if int_delta >= 2**16:
            self.lg.error('{} RNG Ints: Provide int_delta as a 16-bit integer'.format(self.dev_name))
            return None
        if int_delta == 0:
            self.lg.error('{} RNG Ints: Provide int_delta > 0'.format(self.dev_name))
            return None

        comm = 'RNG_INT16S'
        if self.snd_rava_msg(comm, [n_ints, int_delta], 'HH'):
            ints_bytes = await self.get_queue_data(comm)
            if ints_bytes is None:
                return None
            else:
                if output_type == 'list':
                    return bytes_to_list(ints_bytes, 'H')
                else:
                    return bytes_to_array(ints_bytes, np.uint16)


    async def get_rng_floats(self, n_floats, output_type='array'):
        if n_floats == 0:
            self.lg.error('{} RNG Floats: Provide n_floats > 0'.format(self.dev_name))
            return None
        if n_floats >= 2**16:
            self.lg.error('{} RNG Floats: Provide n_floats as a 16-bit integer'.format(self.dev_name))
            return None

        comm = 'RNG_FLOATS'
        if self.snd_rava_msg(comm, [n_floats], 'H'):
            ints_bytes = await self.get_queue_data(comm)
            if ints_bytes is None:
                return None
            else:
                if output_type == 'list':
                    return bytes_to_list(ints_bytes, 'f')
                else:
                    return bytes_to_array(ints_bytes, np.float32)


    async def get_rng_byte_stream_data(self, output_type='array'):
        if not self.rng_streaming:
            self.lg.warning('{} RNG Stream: Streaming is disabled'.format(self.dev_name))
            return None, None

        comm = 'RNG_STREAM_BYTES'
        rng_bytes_a, rng_bytes_b = await self.get_queue_data(comm)

        if (rng_bytes_a is None) or (rng_bytes_b is None):
            return None, None

        else:
            if output_type == 'list':
                rng_a = bytes_to_list(rng_bytes_a, 'B')
                rng_b = bytes_to_list(rng_bytes_b, 'B')
                return rng_a, rng_b

            elif output_type == 'array':
                    rng_a = bytes_to_array(rng_bytes_a, np.uint8)
                    rng_b = bytes_to_array(rng_bytes_b, np.uint8)
                    return rng_a, rng_b

            else:
                return rng_bytes_a, rng_bytes_b


    async def get_rng_byte_stream_status(self):
        comm = 'RNG_STREAM_STATUS'
        if self.snd_rava_msg(comm):
            return await self.get_queue_data(comm)


    ## HEALTH

    async def get_health_startup_results(self):
        comm = 'HEALTH_STARTUP_RESULTS'
        if self.snd_rava_msg(comm):
            success, pc_avg_vars, pc_avg_diff_vars, bias_bit_vars, bias_byte_vars = await self.get_queue_data(comm)
            data_vars = [pc_avg_vars, pc_avg_diff_vars, bias_bit_vars, bias_byte_vars]
            data_names = ['pulse_count', 'pulse_count_diff', 'bit_bias', 'byte_bias']
            return success, dict(zip(data_names, data_vars))


    async def get_health_continuous_errors(self):
        comm = 'HEALTH_CONTINUOUS_ERRORS'
        if self.snd_rava_msg(comm):
            data_vars = await self.get_queue_data(comm)
            n_errors = sum(data_vars)
            data_names = ['repetitive_count_a', 'repetitive_count_b',
                          'adaptative_proportion_a', 'adaptative_proportion_b']
            return n_errors, dict(zip(data_names, data_vars))


    ## PERIPHERALS

    async def get_periph_digi_state(self, periph_id=1):
        if periph_id == 0 or periph_id > 5:
            self.lg.error('{} Periph: Provide a periph_id between 1 and 5'.format(self.dev_name))
            return None

        comm = 'PERIPH_READ'
        if self.snd_rava_msg(comm, [periph_id], 'B'):
            digi_state, digi_mode = await self.get_queue_data(comm)
            return digi_state, D_PERIPH_MODES_INV[digi_mode]


    async def get_periph_d2_input_capture(self):
        comm = 'PERIPH_D2_TIMER3_INPUT_CAPTURE'
        return await self.get_queue_data(comm)


    async def get_periph_d5_adc_read(self, ref_5v=1, clk_prescaler=0, oversampling_n_bits=0, on=True):
        if clk_prescaler == 0 or clk_prescaler > 7:
            self.lg.error('{} Periph D5: Provide a clk_prescaler between 1 and 7'.format(self.dev_name))
            return None
        if oversampling_n_bits > 6:
            self.lg.error('{} Periph D5: Provide a oversampling_n_bits <= 6'.format(self.dev_name))
            return None

        comm = 'PERIPH_D5_ADC'
        if self.snd_rava_msg(comm, [on, ref_5v, clk_prescaler, oversampling_n_bits], 'BBBB'):
            if on:
                return await self.get_queue_data(comm)


    ## INTERFACES

    async def get_interface_ds18bs0(self):
        comm = 'INTERFACE_DS18B20'
        if self.snd_rava_msg(comm):
            return await self.get_queue_data(comm)


"""
Copyright (c) 2024 Gabriel Guerrer

Distributed under the MIT license - See LICENSE for details
"""

"""
This code evokes the RAVA's Tk Acquisition application.

To execute it, enter the following command: python3 -m rng_rava.tk.acq
"""

import rng_rava as rava 

TITLE = 'RAVA Acquisition'

rava_subapp_control_panel = rava.tk.ctrlp.rava_subapp_control_panel
rava_subapp_control_panel['show_button']= False
SUBAPPS = [rava_subapp_control_panel, 
           rava.tk.acq.rava_subapp_acquisition]


def main():
    # RAVA main app
    tkapp = rava.tk.RAVA_APP(title=TITLE, subapp_dicts=SUBAPPS, rava_class=rava.RAVA_RNG, cfg_log_name='rava_acq')

    # Enter Tk loop
    tkapp.mainloop()


if __name__ == '__main__':
    main()


'''
This example showcases asynchronous RNG functionality.

This example code is in the public domain.
Author: Gabriel Guerrer
'''

import asyncio
import rng_rava as rava

async def main():
    # Create a RAVA instance and set the logging level to DEBUG
    rng = rava.RAVA_RNG_AIO()
    rng.log_level('DEBUG')

    # Find RAVA device and connect
    dev_sns = rava.find_rava_sns()
    if len(dev_sns):
        await rng.connect(dev_sns[0])
    else:
        print('No device found')
        exit()

    # Request configuration
    print('\nPWM setup: {}\n'.format(await rng.get_pwm_setup()))
    print('\nRNG setup: {}\n'.format(await rng.get_rng_setup()))

    # Generate random data
    N_DATA = 30
    results = await asyncio.gather(
        rng.get_rng_pulse_counts(n_counts=N_DATA, output_type='list'),
        rng.get_rng_bits(bit_source_id=rava.D_RNG_BIT_SRC['AB']),
        rng.get_rng_bytes(n_bytes=N_DATA, postproc_id=rava.D_RNG_POSTPROC['NONE'], output_type='list'),
        rng.get_rng_int8s(n_ints=N_DATA, int_delta=100, output_type='list'),
        rng.get_rng_int16s(n_ints=N_DATA, int_delta=1000, output_type='list'),
        rng.get_rng_floats(n_floats=N_DATA, output_type='list')
    )

    # Print results
    print('\nRESULTS')
    print('\nPulse Count = {}'.format(results[0]))
    print('\nBits = {}'.format(results[1]))
    print('\nBytes = {}'.format(results[2]))
    print('\nInt8s = {}'.format(results[3]))
    print('\nInt16s = {}'.format(results[4]))
    print('\nFloats = {}'.format(results[5]))

    # Close device
    rng.close()

asyncio.run(main())


"""
Copyright (c) 2023 Gabriel Guerrer

Distributed under the MIT license - See LICENSE for details
"""

"""
The Control Panel sub-application provides an graphical user interface (GUI) for
testing the core features of the RAVA circuit, including EEPROM, PWM, RNG,
HEALTH, LED, and LAMP capabilities.
"""

import tkinter as tk
from tkinter import ttk
import tkinter.messagebox as tkm

from rng_rava.rava_defs import *
from rng_rava import RAVA_RNG_LED
from rng_rava.tk import RAVA_SUBAPP

### VARS

PAD = 10
CTRL_PANEL_N_MAX = 200


### RAVA_SUBAPP_CTRL_PANEL

class RAVA_SUBAPP_CTRL_PANEL(RAVA_SUBAPP):

    cfg_default_str = ''

    def __init__(self, parent):
        ## Initialize RAVA_SUBAPP
        name = 'RAVA_SUBAPP CTRL_PANEL'
        win_title = 'Control Panel'
        win_geometry = '620x350'
        win_resizable = False
        if not super().__init__(parent, name=name, win_title=win_title, win_geometry=win_geometry, win_resizable=win_resizable):
            return
        
        # Led and Lamp enabled?
        self.led_enabled = True if (isinstance(self.rng, RAVA_RNG_LED) and self.rng.led_enabled) else False
        self.lamp_enabled = True if (self.led_enabled and self.rng.lamp_enabled) else False
        
        # Widgets
        self.nb = ttk.Notebook(self, padding=PAD)
        self.nb.grid(row=0, column=0, sticky='nsew')
        self.nb.bind('<<NotebookTabChanged>>', self.notebook_tab_change)

        self.frm_eeprom = ttk.Frame(self, name='eeprom')
        self.frm_eeprom.grid(row=0, column=0, sticky='nsew')
        self.widgets_eeprom()

        self.frm_rng = ttk.Frame(self, name='rng')
        self.frm_rng.grid(row=0, column=0, sticky='nsew')
        self.widgets_rng()

        self.frm_health = ttk.Frame(self, name='health')
        self.frm_health.grid(row=0, column=0, sticky='nsew')
        self.widgets_health()

        self.frm_periph = ttk.Frame(self, name='periph')
        self.frm_periph.grid(row=0, column=0, sticky='nsew')
        self.widgets_periph()

        self.frm_led = ttk.Frame(self, name='led')
        self.frm_led.grid(row=0, column=0, sticky='nsew')
        self.widgets_led()

        self.frm_lamp = ttk.Frame(self, name='lamp')
        self.frm_lamp.grid(row=0, column=0, sticky='nsew')
        self.widgets_lamp()

        self.nb.add(self.frm_eeprom, text=' EEPROM ')
        self.nb.add(self.frm_rng, text=' RNG ')
        self.nb.add(self.frm_health, text=' HEALTH ')
        self.nb.add(self.frm_periph, text=' PERIPH ')
        self.nb.add(self.frm_led, text=' LED ')
        self.nb.add(self.frm_lamp, text=' LAMP ')

        # Key Binds
        self.bind('<Control-Key-q>', lambda event=None: self.master.app_close())

        ## Start
        self.widgets_state()


    def close(self):
        if self.rng.connected():
            # Stop bytes stream
            if self.rng.rng_streaming:
                self.rng_gen_byte_stream_stop()

            # Turn LED off
            if self.led_enabled and self.rng.led_intensity:
                self.rng.snd_led_intensity_fade(0, 1000)

            # LAMP mode off
            if self.lamp_enabled and self.rng.lamp_mode:
                self.rng.snd_lamp_mode(False)

        # Close RAVA_SUBAPP
        super().close()


    def widgets_eeprom(self):
        self.frm_eeprom.columnconfigure([0,1,2], weight=1)
        self.frm_eeprom.rowconfigure([0,1], weight=1)

        # firmware
        self.lbf_eeprom_firmware = ttk.Labelframe(self.frm_eeprom, text='Firmware', padding=(PAD, PAD))
        self.lbf_eeprom_firmware.grid(row=0, column=0, rowspan=2, sticky='nsew', padx=PAD, pady=PAD)
        self.lbf_eeprom_firmware.columnconfigure([0,1], weight=1)
        self.lbf_eeprom_firmware.rowconfigure([0,2], weight=1)

        self.lb_eeprom_firmw_version = ttk.Label(self.lbf_eeprom_firmware, text='Version')
        self.lb_eeprom_firmw_version.grid(row=0, column=0, sticky='w')

        self.var_eeprom_firmw_version = tk.StringVar()
        self.en_eeprom_firmw_version = ttk.Entry(self.lbf_eeprom_firmware, textvariable=self.var_eeprom_firmw_version, width=7)
        self.en_eeprom_firmw_version.state(['readonly'])
        self.en_eeprom_firmw_version.grid(row=0, column=1)

        self.lb_eeprom_firmw_modules = ttk.Label(self.lbf_eeprom_firmware, text='Enabled')
        self.lb_eeprom_firmw_modules.grid(row=1, column=0, columnspan=2, sticky='w')

        self.var_eeprom_firmw_modules = tk.StringVar(value=[''])
        self.lbox_eeprom_firmw_modules = tk.Listbox(self.lbf_eeprom_firmware, listvariable=self.var_eeprom_firmw_modules, selectmode='browse', width=16, height=7)
        self.lbox_eeprom_firmw_modules.bind("<<ListboxSelect>>", None)
        self.lbox_eeprom_firmw_modules.grid(row=2, column=0, columnspan=2)

        # pwm
        self.lbf_eeprom_pwm = ttk.Labelframe(self.frm_eeprom, text='PWM', padding=(PAD, PAD))
        self.lbf_eeprom_pwm.grid(row=0, column=1, sticky='nsew', padx=0, pady=PAD)
        self.lbf_eeprom_pwm.columnconfigure([0,1], weight=1)
        self.lbf_eeprom_pwm.rowconfigure([0,1], weight=1)

        self.lb_eeprom_pwm_freq = ttk.Label(self.lbf_eeprom_pwm, text='Frequency')
        self.lb_eeprom_pwm_freq.grid(row=0, column=0, sticky='w')

        self.cbb_eeprom_pwm_freq = ttk.Combobox(self.lbf_eeprom_pwm, width=7)
        self.cbb_eeprom_pwm_freq.grid(row=0, column=1)
        self.cbb_eeprom_pwm_freq.state(['readonly'])
        self.cbb_eeprom_pwm_freq['values'] = list(D_PWM_FREQ.keys())

        self.lb_eeprom_pwm_duty = ttk.Label(self.lbf_eeprom_pwm, text='Duty Cycle')
        self.lb_eeprom_pwm_duty.grid(row=1, column=0, sticky='w')

        self.var_eeprom_pwm_duty = tk.IntVar()
        self.spb_eeprom_pwm_duty = ttk.Spinbox(self.lbf_eeprom_pwm, from_=1, to=255, increment=1, textvariable=self.var_eeprom_pwm_duty, width=4)
        self.spb_eeprom_pwm_duty.grid(row=1, column=1)

        # rng
        self.lbf_eeprom_rng = ttk.Labelframe(self.frm_eeprom, text='RNG', padding=(PAD, PAD))
        self.lbf_eeprom_rng.grid(row=1, column=1, sticky='nswe', padx=0, pady=(0, PAD))
        self.lbf_eeprom_rng.columnconfigure([0,1], weight=1)
        self.lbf_eeprom_rng.rowconfigure([0], weight=1)

        self.lb_eeprom_rng_si = ttk.Label(self.lbf_eeprom_rng, text='Sampling \ninterval (us)')
        self.lb_eeprom_rng_si.grid(row=0, column=0, sticky='w')

        self.var_eeprom_rng_si = tk.IntVar()
        self.spb_eeprom_rng_si = ttk.Spinbox(self.lbf_eeprom_rng, from_=1, to=255, increment=1, textvariable=self.var_eeprom_rng_si, width=4)
        self.spb_eeprom_rng_si.grid(row=0, column=1)

        # led
        self.lbf_eeprom_led = ttk.Labelframe(self.frm_eeprom, text='LED', padding=(PAD, PAD))
        self.lbf_eeprom_led.grid(row=0, column=2, sticky='nsew', padx=PAD, pady=PAD)
        self.lbf_eeprom_led.columnconfigure([0,1], weight=1)
        self.lbf_eeprom_led.rowconfigure([0], weight=1)

        self.lb_eeprom_led_attached = ttk.Label(self.lbf_eeprom_led, text='Attached')
        self.lb_eeprom_led_attached.grid(row=0, column=0)

        self.cbb_eeprom_led_attached = ttk.Combobox(self.lbf_eeprom_led, width=7)
        self.cbb_eeprom_led_attached.grid(row=0, column=1)
        self.cbb_eeprom_led_attached['values'] = [True, False]
        self.cbb_eeprom_led_attached.state(['readonly'])

        # lamp
        self.lbf_eeprom_lamp = ttk.Labelframe(self.frm_eeprom, text='LAMP', padding=(PAD, PAD))
        self.lbf_eeprom_lamp.grid(row=1, column=2, sticky='nsew', padx=PAD, pady=(0, PAD))
        self.lbf_eeprom_lamp.columnconfigure([0,1], weight=1)
        self.lbf_eeprom_lamp.rowconfigure([0,1,2], weight=1)

        self.lb_eeprom_lamp_exp_dur = ttk.Label(self.lbf_eeprom_lamp, text='Exp. Duration (s)')
        self.lb_eeprom_lamp_exp_dur.grid(row=0, column=0, sticky='w')

        self.var_eeprom_lamp_exp_dur = tk.IntVar()
        self.spb_eeprom_lamp_exp_dur = ttk.Spinbox(self.lbf_eeprom_lamp, from_=60, to=6000, increment=1, textvariable=self.var_eeprom_lamp_exp_dur, width=5)
        self.spb_eeprom_lamp_exp_dur.grid(row=0, column=1)

        self.lb_eeprom_lamp_z_sig = ttk.Label(self.lbf_eeprom_lamp, text='Z-score sig.')
        self.lb_eeprom_lamp_z_sig.grid(row=1, column=0, sticky='w')

        self.var_eeprom_lamp_z_sig = tk.DoubleVar()
        self.spb_eeprom_lamp_z_sig = ttk.Spinbox(self.lbf_eeprom_lamp, from_=1, to=5, increment=0.05, textvariable=self.var_eeprom_lamp_z_sig, width=5)
        self.spb_eeprom_lamp_z_sig.grid(row=1, column=1)

        self.lb_eeprom_lamp_mag_smooth = ttk.Label(self.lbf_eeprom_lamp, text='Mag. smooth')
        self.lb_eeprom_lamp_mag_smooth.grid(row=2, column=0, sticky='w')

        self.var_eeprom_lamp_mag_smooth = tk.IntVar()
        self.spb_eeprom_lamp_mag_smooth = ttk.Spinbox(self.lbf_eeprom_lamp, from_=1, to=1000, increment=1, textvariable=self.var_eeprom_lamp_mag_smooth, width=5)
        self.spb_eeprom_lamp_mag_smooth.grid(row=2, column=1)

        # buttons
        self.frm_eeprom_buttons = ttk.Frame(self.frm_eeprom, padding=(PAD, 0, PAD, PAD))
        self.frm_eeprom_buttons.grid(row=2, column=0, columnspan=3, sticky='nsew')
        self.frm_eeprom_buttons.columnconfigure([0,1,2], weight=1)
        self.frm_eeprom_buttons.rowconfigure([0], weight=1)

        self.bt_eeprom_get = ttk.Button(self.frm_eeprom_buttons, text='Get', command=self.eeprom_get)
        self.bt_eeprom_get.grid(row=0, column=0)

        self.bt_eeprom_set = ttk.Button(self.frm_eeprom_buttons, text='Set', command=self.eeprom_set)
        self.bt_eeprom_set.grid(row=0, column=1)

        self.bt_eeprom_defaults = ttk.Button(self.frm_eeprom_buttons, text='Set Defaults', width=13, command=self.eeprom_set_defaults)
        self.bt_eeprom_defaults.grid(row=0, column=2)


    def widgets_rng(self):
        self.frm_rng.columnconfigure([0,1], weight=1)
        self.frm_rng.rowconfigure([0], weight=1)
        self.frm_rng.rowconfigure([1], weight=2)

        # Setup
        self.lbf_rng_setup = ttk.Labelframe(self.frm_rng, text='Setup', padding=(PAD, PAD))
        self.lbf_rng_setup.grid(row=0, column=0, sticky='nsew', padx=PAD, pady=PAD)
        self.lbf_rng_setup.columnconfigure([0,1,2,3], weight=1)
        self.lbf_rng_setup.rowconfigure([0,1], weight=1)

        self.lb_pwm_freq = ttk.Label(self.lbf_rng_setup, text='PWM Freq.')
        self.lb_pwm_freq.grid(row=0, column=0, sticky='w')

        self.cbb_pwm_freq = ttk.Combobox(self.lbf_rng_setup, width=7)
        self.cbb_pwm_freq.grid(row=0, column=1)
        self.cbb_pwm_freq.state(['readonly'])
        self.cbb_pwm_freq['values'] = list(D_PWM_FREQ.keys())

        self.lb_pwm_duty = ttk.Label(self.lbf_rng_setup, text='PWM Duty')
        self.lb_pwm_duty.grid(row=0, column=2)

        self.var_pwm_duty = tk.IntVar()
        self.spb_pwm_duty = ttk.Spinbox(self.lbf_rng_setup, from_=1, to=255, increment=1, textvariable=self.var_pwm_duty, width=4)
        self.spb_pwm_duty.grid(row=0, column=3)

        self.lb_rng_si = ttk.Label(self.lbf_rng_setup, text='Sampling \ninterval (us)')
        self.lb_rng_si.grid(row=1, column=0, sticky='w')

        self.var_rng_si = tk.IntVar()
        self.spb_rng_si = ttk.Spinbox(self.lbf_rng_setup, from_=1, to=255, increment=1, textvariable=self.var_rng_si, width=4)
        self.spb_rng_si.grid(row=1, column=1)

        self.bt_rng_setup_get = ttk.Button(self.lbf_rng_setup, text='Get', width=6, command=self.rng_setup_get)
        self.bt_rng_setup_get.grid(row=1, column=2)

        self.bt_rng_setup_set = ttk.Button(self.lbf_rng_setup, text='Set', width=6, command=self.rng_setup_set)
        self.bt_rng_setup_set.grid(row=1, column=3)

        # Output
        self.lbf_eeprom_rng_out = ttk.Labelframe(self.frm_rng, text='Output: RNG_A, RNG_B', padding=(PAD, PAD))
        self.lbf_eeprom_rng_out.grid(row=0, column=1, rowspan=2, sticky='nsew', padx=(0, PAD), pady=PAD)
        self.lbf_eeprom_rng_out.columnconfigure([0], weight=1)
        self.lbf_eeprom_rng_out.rowconfigure([0, 1], weight=1)

        self.txt_rng_a = tk.Text(self.lbf_eeprom_rng_out, width=18, state='disabled', wrap='word')
        self.txt_rng_a.grid(row=0, column=0, sticky='nsew')

        self.scroll_y_rng_a = ttk.Scrollbar(self.lbf_eeprom_rng_out, orient=tk.VERTICAL, command=self.txt_rng_a.yview)
        self.scroll_y_rng_a.grid(row=0, column=1, sticky='ns')
        self.txt_rng_a['yscrollcommand'] = self.scroll_y_rng_a.set

        self.txt_rng_b = tk.Text(self.lbf_eeprom_rng_out, width=18, state='disabled', wrap='word')
        self.txt_rng_b.grid(row=1, column=0, sticky='nsew', pady=(PAD, 0))

        self.scroll_y_rng_b = ttk.Scrollbar(self.lbf_eeprom_rng_out, orient=tk.VERTICAL, command=self.txt_rng_b.yview)
        self.scroll_y_rng_b.grid(row=1, column=1, sticky='ns', pady=(PAD, 0))
        self.txt_rng_b['yscrollcommand'] = self.scroll_y_rng_b.set

        ## Notebook
        self.nb_rng = ttk.Notebook(self.frm_rng)
        self.nb_rng.grid(row=1, column=0, sticky='nsew', padx=PAD, pady=(0, PAD))
        self.nb_rng.bind('<<NotebookTabChanged>>', self.rng_notebook_tab_change)

        # Data
        self.frm_rng_data = ttk.Frame(self.frm_rng, name='data', padding=PAD)
        self.frm_rng_data.grid(row=0, column=0, sticky='nsew')
        self.frm_rng_data.columnconfigure([0,1,2], weight=1)
        self.frm_rng_data.rowconfigure([0,1,2], weight=1)

        self.lb_rng_data_ammount = ttk.Label(self.frm_rng_data, text='Ammount N')
        self.lb_rng_data_ammount.grid(row=0, column=0, sticky='w')

        self.var_rng_data_ammount = tk.IntVar(value=10)
        self.spb_rng_data_ammount = ttk.Spinbox(self.frm_rng_data, from_=1, to=CTRL_PANEL_N_MAX, increment=5, textvariable=self.var_rng_data_ammount, width=5)
        self.spb_rng_data_ammount.grid(row=0, column=1)

        self.bt_rng_data_pulsecount = ttk.Button(self.frm_rng_data, text='N Pulse Counts', width=13, command=self.rng_gen_pulse_count)
        self.bt_rng_data_pulsecount.grid(row=0, column=2)

        self.lb_rng_data_bit_src = ttk.Label(self.frm_rng_data, text='Bit Source')
        self.lb_rng_data_bit_src.grid(row=1, column=0, sticky='w')

        self.cbb_rng_data_bit_src = ttk.Combobox(self.frm_rng_data, width=9)
        self.cbb_rng_data_bit_src.grid(row=1, column=1)
        self.cbb_rng_data_bit_src.state(['readonly'])
        self.cbb_rng_data_bit_src['values'] = list(D_RNG_BIT_SRC.keys())
        self.cbb_rng_data_bit_src.set(list(D_RNG_BIT_SRC.keys())[0])

        self.bt_rng_data_bit = ttk.Button(self.frm_rng_data, text='1 Bit', width=13, command=self.rng_gen_bit)
        self.bt_rng_data_bit.grid(row=1, column=2)

        self.lb_rng_data_byte_pp = ttk.Label(self.frm_rng_data, text='Post Processing')
        self.lb_rng_data_byte_pp.grid(row=2, column=0, sticky='w')

        self.cbb_rng_data_byte_pp = ttk.Combobox(self.frm_rng_data, width=9)
        self.cbb_rng_data_byte_pp.grid(row=2, column=1)
        self.cbb_rng_data_byte_pp.state(['readonly'])
        self.cbb_rng_data_byte_pp['values'] = list(D_RNG_POSTPROC.keys())
        self.cbb_rng_data_byte_pp.set(list(D_RNG_POSTPROC.keys())[0])

        self.bt_rng_data_byte = ttk.Button(self.frm_rng_data, text='N Bytes', width=13, command=self.rng_gen_byte)
        self.bt_rng_data_byte.grid(row=2, column=2)

        # Numbers
        self.frm_rng_num = ttk.Frame(self.frm_rng, name='num', padding=PAD)
        self.frm_rng_num.grid(row=0, column=0, sticky='nsew')
        self.frm_rng_num.columnconfigure([0,1,2,3,4], weight=1)
        self.frm_rng_num.rowconfigure([0,1,2], weight=1)

        self.lb_rng_num_ammount = ttk.Label(self.frm_rng_num, text='Ammount N')
        self.lb_rng_num_ammount.grid(row=0, column=0, columnspan=2, sticky='w')

        self.var_rng_num_ammount = tk.IntVar(value=10)
        self.spb_rng_num_ammount = ttk.Spinbox(self.frm_rng_num, from_=1, to=100, increment=5, textvariable=self.var_rng_num_ammount, width=5)
        self.spb_rng_num_ammount.grid(row=0, column=2)

        self.lb_rng_num_min_int = ttk.Label(self.frm_rng_num, text='Min')
        self.lb_rng_num_min_int.grid(row=1, column=0, sticky='w')

        self.var_rng_num_min_int = tk.IntVar(value=0)
        self.spb_rng_num_min_int = ttk.Spinbox(self.frm_rng_num, from_=0, to=65535, increment=5, textvariable=self.var_rng_num_min_int, width=4)
        self.spb_rng_num_min_int.grid(row=1, column=1)

        self.lb_rng_num_max_int = ttk.Label(self.frm_rng_num, text='Max')
        self.lb_rng_num_max_int.grid(row=1, column=2)

        self.var_rng_num_max_int = tk.IntVar(value=9)
        self.spb_rng_num_max_int = ttk.Spinbox(self.frm_rng_num, from_=1, to=65535, increment=5, textvariable=self.var_rng_num_max_int, width=4)
        self.spb_rng_num_max_int.grid(row=1, column=3)

        self.bt_rng_num_int = ttk.Button(self.frm_rng_num, text='N Integers', width=10, command=self.rng_gen_integer)
        self.bt_rng_num_int.grid(row=1, column=4)

        self.lb_rng_num_min_float = ttk.Label(self.frm_rng_num, text='Min')
        self.lb_rng_num_min_float.grid(row=2, column=0, sticky='w')

        self.var_rng_num_min_float = tk.DoubleVar(value=0.)
        self.spb_rng_num_min_float = ttk.Spinbox(self.frm_rng_num, from_=0, to=1000, increment=5, textvariable=self.var_rng_num_min_float, width=4)
        self.spb_rng_num_min_float.grid(row=2, column=1)

        self.lb_rng_num_max_float = ttk.Label(self.frm_rng_num, text='Max')
        self.lb_rng_num_max_float.grid(row=2, column=2)

        self.var_rng_num_max_float = tk.DoubleVar(value=1.)
        self.spb_rng_num_max_float = ttk.Spinbox(self.frm_rng_num, from_=0, to=1000, increment=0.5, textvariable=self.var_rng_num_max_float, width=4)
        self.spb_rng_num_max_float.grid(row=2, column=3)

        self.bt_rng_num_float = ttk.Button(self.frm_rng_num, text='N Floats', width=10, command=self.rng_gen_float)
        self.bt_rng_num_float.grid(row=2, column=4)

        # Byte stream
        self.frm_rng_stream = ttk.Frame(self.frm_rng, name='stream', padding=PAD)
        self.frm_rng_stream.grid(row=0, column=0, sticky='nsew')
        self.frm_rng_stream.columnconfigure([0,1,2,3], weight=1)
        self.frm_rng_stream.rowconfigure([0,1,2], weight=1)

        self.lb_rng_stream_ammount = ttk.Label(self.frm_rng_stream, text='Bytes per stream')
        self.lb_rng_stream_ammount.grid(row=0, column=0, columnspan=2, sticky='w')

        self.var_rng_stream_ammount = tk.IntVar(value=10)
        self.spb_rng_stream_ammount = ttk.Spinbox(self.frm_rng_stream, from_=1, to=CTRL_PANEL_N_MAX, increment=5, textvariable=self.var_rng_stream_ammount, width=5)
        self.spb_rng_stream_ammount.grid(row=0, column=1, columnspan=2)

        self.lb_rng_stream_interval = ttk.Label(self.frm_rng_stream, text='Stream \ninterval (ms)')
        self.lb_rng_stream_interval.grid(row=1, column=0, sticky='w')

        self.var_rng_stream_interval = tk.IntVar(value=200)
        self.spb_rng_stream_interval = ttk.Spinbox(self.frm_rng_stream, from_=1, to=RNG_BYTE_STREAM_MAX_INTERVAL_MS, increment=100, textvariable=self.var_rng_stream_interval, width=4)
        self.spb_rng_stream_interval.grid(row=1, column=1)

        self.lb_rng_stream_byte_pp = ttk.Label(self.frm_rng_stream, text='Post Proc.')
        self.lb_rng_stream_byte_pp.grid(row=1, column=2)

        self.cbb_rng_stream_byte_pp = ttk.Combobox(self.frm_rng_stream, width=7)
        self.cbb_rng_stream_byte_pp.grid(row=1, column=3)
        self.cbb_rng_stream_byte_pp.state(['readonly'])
        self.cbb_rng_stream_byte_pp['values'] = list(D_RNG_POSTPROC.keys())
        self.cbb_rng_stream_byte_pp.set(list(D_RNG_POSTPROC.keys())[0])

        self.bt_rng_stream_start = ttk.Button(self.frm_rng_stream, text='Start / Get', width=10, command=self.rng_gen_byte_stream_start)
        self.bt_rng_stream_start.grid(row=2, column=0, columnspan=2)

        self.bt_rng_stream_stop = ttk.Button(self.frm_rng_stream, text='Stop', width=10, command=self.rng_gen_byte_stream_stop)
        self.bt_rng_stream_stop.grid(row=2, column=2, columnspan=2)

        # Notebook
        self.nb_rng.add(self.frm_rng_data, text='Data ')
        self.nb_rng.add(self.frm_rng_num, text='Numbers ')
        self.nb_rng.add(self.frm_rng_stream, text='Byte Stream ')


    def widgets_health(self):
        self.frm_health.columnconfigure([0,1], weight=1)
        self.frm_health.rowconfigure([0,1,2], weight=1)

        # Startup
        self.lbf_health_startup = ttk.Labelframe(self.frm_health, text='Startup Tests', padding=(PAD, PAD))
        self.lbf_health_startup.grid(row=0, column=0, sticky='nsew', padx=PAD, pady=PAD)
        self.lbf_health_startup.columnconfigure([0], weight=1)
        self.lbf_health_startup.rowconfigure([0,1], weight=1)

        self.bt_rng_health_startup_run = ttk.Button(self.lbf_health_startup, text='Run', width=10, command=self.health_startup_run)
        self.bt_rng_health_startup_run.grid(row=0, column=0)

        self.bt_rng_health_startup_get = ttk.Button(self.lbf_health_startup, text='Get', width=10, command=self.health_startup_get)
        self.bt_rng_health_startup_get.grid(row=1, column=0)

        # Continuous
        self.lbf_health_cont = ttk.Labelframe(self.frm_health, text='Continuous Tests', padding=(PAD, PAD))
        self.lbf_health_cont.grid(row=1, column=0, sticky='nsew', padx=PAD, pady=(0, PAD))
        self.lbf_health_cont.columnconfigure([0], weight=1)
        self.lbf_health_cont.rowconfigure([0], weight=1)

        self.bt_health_cont_get = ttk.Button(self.lbf_health_cont, text='Get', width=10, command=self.health_continuous_get)
        self.bt_health_cont_get.grid(row=0, column=0)

        # Results
        self.lbf_health_results = ttk.Labelframe(self.frm_health, text='Results', padding=(PAD, PAD))
        self.lbf_health_results.grid(row=0, column=1, rowspan=3, sticky='nsew', padx=(0, PAD), pady=PAD)
        self.lbf_health_results.columnconfigure([0], weight=1)
        self.lbf_health_results.rowconfigure([0], weight=1)

        self.txt_health_results = tk.Text(self.lbf_health_results, width=50, state='disabled', wrap='word')
        self.txt_health_results.grid(row=0, column=0, sticky='nsew')


    def widgets_periph(self):
        self.frm_periph.columnconfigure([0,1], weight=1)
        self.frm_periph.rowconfigure([0,1], weight=1)

        # Digital IO
        self.lbf_periph_digi = ttk.Labelframe(self.frm_periph, text='Digital IO', padding=(PAD, PAD))
        self.lbf_periph_digi.grid(row=0, column=0, sticky='nsew', padx=PAD, pady=PAD)
        self.lbf_periph_digi.columnconfigure([0,1,2,3], weight=1)
        self.lbf_periph_digi.rowconfigure([0,1,2,3], weight=1)

        self.lb_periph_port = ttk.Label(self.lbf_periph_digi, text='Port')
        self.lb_periph_port.grid(row=0, column=0, sticky='w')

        self.var_periph_port = tk.IntVar(value=1)
        self.cbb_periph_port = ttk.Combobox(self.lbf_periph_digi, width=5, textvariable=self.var_periph_port)
        self.cbb_periph_port.grid(row=0, column=1, sticky='w')
        self.cbb_periph_port.state(['readonly'])
        self.cbb_periph_port['values'] = list(range(1, PERIPH_PORTS + 1))
        self.cbb_periph_port.bind('<<ComboboxSelected>>', self.periph_digi_state_update)

        self.lb_periph_mode = ttk.Label(self.lbf_periph_digi, text='Mode')
        self.lb_periph_mode.grid(row=1, column=0, sticky='w')

        self.cbb_periph_mode = ttk.Combobox(self.lbf_periph_digi, width=8)
        self.cbb_periph_mode.grid(row=1, column=1, sticky='w')
        self.cbb_periph_mode.state(['readonly'])
        self.cbb_periph_mode['values'] = ['INPUT', 'OUTPUT']
        self.cbb_periph_mode.set('INPUT')
        self.cbb_periph_mode.bind('<<ComboboxSelected>>', self.periph_digi_mode)

        self.lb_periph_state = ttk.Label(self.lbf_periph_digi, text='State')
        self.lb_periph_state.grid(row=2, column=0, sticky='w')

        self.var_periph_state = tk.IntVar(value=0)
        self.cbb_periph_state = ttk.Combobox(self.lbf_periph_digi, width=5, textvariable=self.var_periph_state)
        self.cbb_periph_state.grid(row=2, column=1, sticky='w')
        self.cbb_periph_state.state(['readonly'])
        self.cbb_periph_state['values'] = [0, 1]

        self.bt_periph_state_get = ttk.Button(self.lbf_periph_digi, text='Get', width=5, command=self.periph_digi_state_get)
        self.bt_periph_state_get.grid(row=2, column=2)

        self.bt_periph_state_set = ttk.Button(self.lbf_periph_digi, text='Set', width=5, command=self.periph_digi_state_set)
        self.bt_periph_state_set.grid(row=2, column=3)

        self.lb_periph_pulse_dur = ttk.Label(self.lbf_periph_digi, text='Dur. (us)')
        self.lb_periph_pulse_dur.grid(row=3, column=0, sticky='w')

        self.var_periph_pulse_dur = tk.IntVar(value=100)
        self.spb_periph_port = ttk.Spinbox(self.lbf_periph_digi, from_=1, to=100000, increment=1, textvariable=self.var_periph_pulse_dur, width=5)
        self.spb_periph_port.grid(row=3, column=1, sticky='w')

        self.bt_periph_pulse = ttk.Button(self.lbf_periph_digi, text='Pulse', width=7, command=self.periph_digi_pulse)
        self.bt_periph_pulse.grid(row=3, column=2, columnspan=2)

        ## Notebook
        self.nb_periph = ttk.Notebook(self.frm_periph)
        self.nb_periph.grid(row=0, column=1, sticky='nsew', padx=(0, PAD), pady=PAD)

        # D1
        self.frm_periph_d1 = ttk.Frame(self.frm_periph, padding=(PAD, PAD))
        self.frm_periph_d1.grid(row=0, column=0, sticky='nsew', padx=(0, PAD), pady=PAD)
        self.frm_periph_d1.columnconfigure([0,1,2], weight=1)
        self.frm_periph_d1.rowconfigure([0,1,2,3], weight=1)

        self.lb_periph_d1_timing_dbg = ttk.Label(self.frm_periph_d1, text='Timing Debug')
        self.lb_periph_d1_timing_dbg.grid(row=0, column=0, columnspan=2, sticky='w')

        self.cbb_periph_d1_timing_dbg = ttk.Combobox(self.frm_periph_d1, width=5)
        self.cbb_periph_d1_timing_dbg.grid(row=0, column=2)
        self.cbb_periph_d1_timing_dbg.state(['readonly'])
        self.cbb_periph_d1_timing_dbg['values'] = ['On', 'Off']
        self.cbb_periph_d1_timing_dbg.set('Off')
        self.cbb_periph_d1_timing_dbg.bind('<<ComboboxSelected>>', self.periph_d1_timing_debug)

        self.lb_periph_d1_trigger = ttk.Label(self.frm_periph_d1, text='Trigger Input')
        self.lb_periph_d1_trigger.grid(row=1, column=0, columnspan=2, sticky='w')

        self.cbb_periph_d1_trigger_on = ttk.Combobox(self.frm_periph_d1, width=5)
        self.cbb_periph_d1_trigger_on.grid(row=1, column=2)
        self.cbb_periph_d1_trigger_on.state(['readonly'])
        self.cbb_periph_d1_trigger_on['values'] = ['On', 'Off']
        self.cbb_periph_d1_trigger_on.set('Off')
        self.cbb_periph_d1_trigger_on.bind('<<ComboboxSelected>>', self.periph_d1_trigger_input)

        self.lb_periph_d1_comparator = ttk.Label(self.frm_periph_d1, text='Comparator')
        self.lb_periph_d1_comparator.grid(row=2, column=0, sticky='w')

        self.cbb_periph_d1_comparator_neg = ttk.Combobox(self.frm_periph_d1, width=7)
        self.cbb_periph_d1_comparator_neg.grid(row=2, column=1)
        self.cbb_periph_d1_comparator_neg.state(['readonly'])
        self.cbb_periph_d1_comparator_neg['values'] = ['Neg=D5', 'Neg=0']
        self.cbb_periph_d1_comparator_neg.set('Neg=0')
        self.cbb_periph_d1_comparator_neg.bind('<<ComboboxSelected>>', self.periph_d1_comparator)

        self.cbb_periph_d1_comparator_on = ttk.Combobox(self.frm_periph_d1, width=5)
        self.cbb_periph_d1_comparator_on.grid(row=2, column=2)
        self.cbb_periph_d1_comparator_on.state(['readonly'])
        self.cbb_periph_d1_comparator_on['values'] = ['On', 'Off']
        self.cbb_periph_d1_comparator_on.set('Off')
        self.cbb_periph_d1_comparator_on.bind('<<ComboboxSelected>>', self.periph_d1_comparator)

        self.lb_periph_d1_delay = ttk.Label(self.frm_periph_d1, text='Delay Dur. (us)')
        self.lb_periph_d1_delay.grid(row=3, column=0, sticky='w')

        self.var_periph_d1_delay_dur = tk.IntVar(value=10)
        self.spb_periph_d1_delay_dur = ttk.Spinbox(self.frm_periph_d1, from_=1, to=100000, increment=1, textvariable=self.var_periph_d1_delay_dur, width=5)
        self.spb_periph_d1_delay_dur.grid(row=3, column=1)

        self.bt_periph_d1_delay = ttk.Button(self.frm_periph_d1, text='Test', width=7, command=self.periph_d1_delay_test)
        self.bt_periph_d1_delay.grid(row=3, column=2)

        # D2
        self.frm_periph_d2 = ttk.Frame(self.frm_periph, padding=(PAD, PAD))
        self.frm_periph_d2.grid(row=0, column=0, sticky='nsew', padx=PAD, pady=(0, PAD))
        self.frm_periph_d2.columnconfigure([0,1], weight=1)
        self.frm_periph_d2.rowconfigure([0,1], weight=1)

        self.lb_periph_d2_input_capture = ttk.Label(self.frm_periph_d2, text='Input Capture')
        self.lb_periph_d2_input_capture.grid(row=0, column=0)

        self.cbb_periph_d2_input_capture_on = ttk.Combobox(self.frm_periph_d2, width=7)
        self.cbb_periph_d2_input_capture_on.grid(row=0, column=1)
        self.cbb_periph_d2_input_capture_on.state(['readonly'])
        self.cbb_periph_d2_input_capture_on['values'] = ['On', 'Off']
        self.cbb_periph_d2_input_capture_on.set('Off')
        self.cbb_periph_d2_input_capture_on.bind('<<ComboboxSelected>>', self.periph_d2_input_capture)

        self.lb_periph_d2_input_capture_interval = ttk.Label(self.frm_periph_d2, text='Capture Interval (s)')
        self.lb_periph_d2_input_capture_interval.grid(row=1, column=0)

        self.var_periph_d2_input_capture_interval = tk.StringVar()
        self.en_periph_d2_input_capture_interval = ttk.Entry(self.frm_periph_d2, width=10, textvariable=self.var_periph_d2_input_capture_interval)
        self.en_periph_d2_input_capture_interval.grid(row=1, column=1)
        self.en_periph_d2_input_capture_interval.state(['readonly'])

        # D3
        self.frm_periph_d3 = ttk.Frame(self.frm_periph, padding=(PAD, PAD))
        self.frm_periph_d3.grid(row=0, column=0, sticky='nsew', padx=(0, PAD), pady=(0, PAD))
        self.frm_periph_d3.columnconfigure([0,1,2], weight=1)
        self.frm_periph_d3.rowconfigure([0,1,2,3,4], weight=1)

        self.lb_periph_d3_trigger_out_interval = ttk.Label(self.frm_periph_d3, text='Trigger Output\nInterval (ms)')
        self.lb_periph_d3_trigger_out_interval.grid(row=0, column=0, sticky='w')

        self.var_periph_d3_trigger_out_interval = tk.IntVar(value=10)
        self.spb_periph_d3_trigger_out_interval = ttk.Spinbox(self.frm_periph_d3, from_=1, to=100000, increment=1, textvariable=self.var_periph_d3_trigger_out_interval, width=5, command=self.periph_d3_trigger_output)
        self.spb_periph_d3_trigger_out_interval.grid(row=0, column=1)

        self.cbb_periph_d3_trigger_out_on = ttk.Combobox(self.frm_periph_d3, width=5)
        self.cbb_periph_d3_trigger_out_on.grid(row=0, column=2)
        self.cbb_periph_d3_trigger_out_on.state(['readonly'])
        self.cbb_periph_d3_trigger_out_on['values'] = ['On', 'Off']
        self.cbb_periph_d3_trigger_out_on.set('Off')
        self.cbb_periph_d3_trigger_out_on.bind('<<ComboboxSelected>>', self.periph_d3_trigger_output)

        self.lb_periph_d3_pwm = ttk.Label(self.frm_periph_d3, text='PWM Output')
        self.lb_periph_d3_pwm.grid(row=1, column=0, sticky='w')

        self.cbb_periph_d3_pwm_on = ttk.Combobox(self.frm_periph_d3, width=5)
        self.cbb_periph_d3_pwm_on.grid(row=1, column=2)
        self.cbb_periph_d3_pwm_on.state(['readonly'])
        self.cbb_periph_d3_pwm_on['values'] = ['On', 'Off']
        self.cbb_periph_d3_pwm_on.set('Off')
        self.cbb_periph_d3_pwm_on.bind('<<ComboboxSelected>>', self.periph_d3_pwm)

        self.lb_periph_d3_pwm_presc = ttk.Label(self.frm_periph_d3, text='Prescaler')
        self.lb_periph_d3_pwm_presc.grid(row=2, column=1, sticky='w')

        self.var_periph_d3_pwm_presc = tk.IntVar(value=1)
        self.cbb_periph_d3_pwm_presc = ttk.Combobox(self.frm_periph_d3, width=5, textvariable=self.var_periph_d3_pwm_presc)
        self.cbb_periph_d3_pwm_presc.grid(row=2, column=2)
        self.cbb_periph_d3_pwm_presc.state(['readonly'])
        self.cbb_periph_d3_pwm_presc['values'] = list(range(1,5+1))
        self.cbb_periph_d3_pwm_presc.bind('<<ComboboxSelected>>', self.periph_d3_pwm)

        self.lb_periph_d3_pwm_top = ttk.Label(self.frm_periph_d3, text='Top')
        self.lb_periph_d3_pwm_top.grid(row=3, column=1, sticky='w')

        self.var_periph_d3_pwm_top = tk.IntVar(value=65535)
        self.spb_periph_d3_pwm_top = ttk.Spinbox(self.frm_periph_d3, from_=1, to=65535, increment=1000, textvariable=self.var_periph_d3_pwm_top, width=5, command=self.periph_d3_pwm)
        self.spb_periph_d3_pwm_top.grid(row=3, column=2)

        self.lb_periph_d3_pwm_duty = ttk.Label(self.frm_periph_d3, text='Duty (%)')
        self.lb_periph_d3_pwm_duty.grid(row=4, column=1, sticky='w')

        self.var_periph_d3_pwm_duty = tk.DoubleVar(value=10)
        self.spb_periph_d3_pwm_duty = ttk.Spinbox(self.frm_periph_d3, from_=1, to=100, increment=5, textvariable=self.var_periph_d3_pwm_duty, width=5, command=self.periph_d3_pwm)
        self.spb_periph_d3_pwm_duty.grid(row=4, column=2)

        # D4
        self.frm_periph_d4 = ttk.Frame(self.frm_periph, padding=(PAD, PAD))
        self.frm_periph_d4.grid(row=0, column=0, sticky='nsew', padx=PAD, pady=(0, PAD))
        self.frm_periph_d4.columnconfigure([0,1], weight=1)
        self.frm_periph_d4.rowconfigure([0], weight=1)

        self.lb_periph_d4_pin_change = ttk.Label(self.frm_periph_d4, text='Pin Change')
        self.lb_periph_d4_pin_change.grid(row=0, column=0)

        self.cbb_periph_d4_pin_change_on = ttk.Combobox(self.frm_periph_d4, width=7)
        self.cbb_periph_d4_pin_change_on.grid(row=0, column=1)
        self.cbb_periph_d4_pin_change_on.state(['readonly'])
        self.cbb_periph_d4_pin_change_on['values'] = ['On', 'Off']
        self.cbb_periph_d4_pin_change_on.set('Off')
        self.cbb_periph_d4_pin_change_on.bind('<<ComboboxSelected>>', self.periph_d4_pin_change)

        # D5
        self.frm_periph_d5 = ttk.Frame(self.frm_periph, padding=(PAD, PAD))
        self.frm_periph_d5.grid(row=0, column=0, sticky='nsew', padx=PAD, pady=(0, PAD))
        self.frm_periph_d5.columnconfigure([0,1,2,3], weight=1)
        self.frm_periph_d5.rowconfigure([0,1,2,3], weight=1)

        self.lb_periph_d5_adc = ttk.Label(self.frm_periph_d5, text='ADC')
        self.lb_periph_d5_adc.grid(row=0, column=0)

        self.var_periph_d5_adc = tk.StringVar()
        self.en_periph_d5_adc = ttk.Entry(self.frm_periph_d5, textvariable=self.var_periph_d5_adc, width=12)
        self.en_periph_d5_adc.state(['readonly'])
        self.en_periph_d5_adc.grid(row=0, column=1)

        self.bt_periph_d5_adc = ttk.Button(self.frm_periph_d5, text='Get', width=7, command=self.periph_d5_adc_read)
        self.bt_periph_d5_adc.grid(row=0, column=2)

        self.lb_periph_d5_adc_ref = ttk.Label(self.frm_periph_d5, text='Reference')
        self.lb_periph_d5_adc_ref.grid(row=1, column=1)

        self.cbb_periph_d5_adc_ref = ttk.Combobox(self.frm_periph_d5, width=4)
        self.cbb_periph_d5_adc_ref.grid(row=1, column=2)
        self.cbb_periph_d5_adc_ref.state(['readonly'])
        self.cbb_periph_d5_adc_ref['values'] = ['5V', '2.6V']
        self.cbb_periph_d5_adc_ref.set('5V')

        self.lb_periph_d5_adc_prescaler = ttk.Label(self.frm_periph_d5, text='Prescaler')
        self.lb_periph_d5_adc_prescaler.grid(row=2, column=1)

        self.var_periph_d5_adc_prescaler = tk.IntVar(value=6)
        self.cbb_periph_d5_adc_prescaler = ttk.Combobox(self.frm_periph_d5, width=3, textvariable=self.var_periph_d5_adc_prescaler)
        self.cbb_periph_d5_adc_prescaler.grid(row=2, column=2)
        self.cbb_periph_d5_adc_prescaler.state(['readonly'])
        self.cbb_periph_d5_adc_prescaler['values'] = list(range(1, 7+1))

        self.lb_periph_d5_adc_oversampling = ttk.Label(self.frm_periph_d5, text='Oversampling')
        self.lb_periph_d5_adc_oversampling.grid(row=3, column=1)

        self.var_periph_d5_adc_oversampling = tk.IntVar(value=0)
        self.cbb_periph_d5_adc_oversampling = ttk.Combobox(self.frm_periph_d5, width=3, textvariable=self.var_periph_d5_adc_oversampling)
        self.cbb_periph_d5_adc_oversampling.grid(row=3, column=2)
        self.cbb_periph_d5_adc_oversampling.state(['readonly'])
        self.cbb_periph_d5_adc_oversampling['values'] = list(range(0, 6+1))

        # Notebook
        self.nb_periph.add(self.frm_periph_d1, text=' D1 ')
        self.nb_periph.add(self.frm_periph_d2, text=' D2 ')
        self.nb_periph.add(self.frm_periph_d3, text=' D3 ')
        self.nb_periph.add(self.frm_periph_d4, text=' D4 ')
        self.nb_periph.add(self.frm_periph_d5, text=' D5 ')


    def widgets_led(self):
        self.frm_led.columnconfigure([0,1], weight=1)
        self.frm_led.rowconfigure([0,1], weight=1)

        # Color
        self.lbf_led_color = ttk.Labelframe(self.frm_led, text='Color', padding=(PAD, PAD))
        self.lbf_led_color.grid(row=0, column=0, sticky='nsew', padx=PAD, pady=PAD)
        self.lbf_led_color.columnconfigure([0,1,2], weight=1)
        self.lbf_led_color.rowconfigure([0,1,2,3], weight=1)

        self.var_led_color = tk.IntVar(value=0)
        self.scl_led_color = ttk.Scale(self.lbf_led_color, orient=tk.HORIZONTAL, length=170, from_=0, to=255, variable=self.var_led_color, command=self.led_color_scale)
        self.scl_led_color.grid(row=0, column=0, columnspan=2)

        self.var_led_color_str = tk.StringVar(value='RED')
        self.cbb_led_color = ttk.Combobox(self.lbf_led_color, width=7, textvariable=self.var_led_color_str)
        self.cbb_led_color.grid(row=0, column=2)
        self.cbb_led_color.state(['readonly'])
        self.cbb_led_color['values'] = list(D_LED_COLOR.keys())
        self.cbb_led_color.bind('<<ComboboxSelected>>', self.led_color_cbb)

        self.lb_led_color_fade = ttk.Label(self.lbf_led_color, text='Fade Dur. (ms)')
        self.lb_led_color_fade.grid(row=1, column=0)

        self.var_led_color_fade_dur = tk.IntVar(value=1000)
        self.spb_led_color_fade_dur = ttk.Spinbox(self.lbf_led_color, from_=500, to=10000, increment=500, textvariable=self.var_led_color_fade_dur, width=5)
        self.spb_led_color_fade_dur.grid(row=1, column=1)

        self.var_led_color_fade_str = tk.StringVar(value='RED')
        self.cbb_led_color_fade = ttk.Combobox(self.lbf_led_color, width=7, textvariable=self.var_led_color_fade_str)
        self.cbb_led_color_fade.grid(row=1, column=2)
        self.cbb_led_color_fade.state(['readonly'])
        self.cbb_led_color_fade['values'] = list(D_LED_COLOR.keys())
        self.cbb_led_color_fade.bind('<<ComboboxSelected>>', self.led_color_fade)

        self.lb_led_color_osc_cycles = ttk.Label(self.lbf_led_color, text='Osc. Cycles')
        self.lb_led_color_osc_cycles.grid(row=2, column=0)

        self.var_led_color_osc_cycles = tk.IntVar(value=3)
        self.spb_led_color_osc_cycles = ttk.Spinbox(self.lbf_led_color, from_=1, to=10, increment=1, textvariable=self.var_led_color_osc_cycles, width=5)
        self.spb_led_color_osc_cycles.grid(row=2, column=1)

        self.lb_led_color_osc_dur = ttk.Label(self.lbf_led_color, text='Osc. Dur. (ms)')
        self.lb_led_color_osc_dur.grid(row=3, column=0)

        self.var_led_color_osc_dur = tk.IntVar(value=3000)
        self.spb_led_color_osc_dur = ttk.Spinbox(self.lbf_led_color, from_=500, to=10000, increment=500, textvariable=self.var_led_color_osc_dur, width=5)
        self.spb_led_color_osc_dur.grid(row=3, column=1)

        self.bt_led_color_osc = ttk.Button(self.lbf_led_color, text='Oscillate', width=8, command=self.led_color_oscillate)
        self.bt_led_color_osc.grid(row=2, column=2, rowspan=2)

        # Intensity
        self.lbf_led_intensity = ttk.Labelframe(self.frm_led, text='Intensity', padding=(PAD, PAD))
        self.lbf_led_intensity.grid(row=0, column=1, sticky='nsew', padx=(0, PAD), pady=PAD)
        self.lbf_led_intensity.columnconfigure([0,1], weight=1)
        self.lbf_led_intensity.rowconfigure([0,1,2,3], weight=1)

        self.var_led_int_scale = tk.IntVar(value=0)
        self.scl_led_int = ttk.Scale(self.lbf_led_intensity, orient=tk.HORIZONTAL, length=170, from_=0, to=255, variable=self.var_led_int_scale, command=self.led_intensity_scale)
        self.scl_led_int.grid(row=0, column=0, columnspan=2)

        self.var_led_int_spb = tk.IntVar(value=0)
        self.spb_led_int = ttk.Spinbox(self.lbf_led_intensity, from_=0, to=255, increment=10, textvariable=self.var_led_int_spb, width=5, command=self.led_intensity_spb)
        self.spb_led_int.grid(row=0, column=2)

        self.lb_led_int_fade = ttk.Label(self.lbf_led_intensity, text='Fade Dur. (ms)')
        self.lb_led_int_fade.grid(row=1, column=0)

        self.var_led_int_fade_dur = tk.IntVar(value=1000)
        self.spb_led_int_fade_dur = ttk.Spinbox(self.lbf_led_intensity, from_=500, to=10000, increment=500, textvariable=self.var_led_int_fade_dur, width=5)
        self.spb_led_int_fade_dur.grid(row=1, column=1)

        self.var_led_int_fade = tk.IntVar(value=0)
        self.cbb_led_int_fade = ttk.Combobox(self.lbf_led_intensity, width=7, textvariable=self.var_led_int_fade)
        self.cbb_led_int_fade.grid(row=1, column=2)
        self.cbb_led_int_fade.state(['readonly'])
        self.cbb_led_int_fade['values'] = [0, 64, 128, 192, 255]
        self.cbb_led_int_fade.bind('<<ComboboxSelected>>', self.led_intensity_fade)

        self.lb_led_int_spacer1 = ttk.Label(self.lbf_led_intensity, text=' ')
        self.lb_led_int_spacer1.grid(row=2, column=0)

        self.lb_led_int_spacer2 = ttk.Label(self.lbf_led_intensity, text=' ')
        self.lb_led_int_spacer2.grid(row=3, column=0)


    def widgets_lamp(self):
        self.frm_lamp.columnconfigure([0,1], weight=1)
        self.frm_lamp.rowconfigure([0,1], weight=1)

        # Commands
        self.lbf_lamp_comm = ttk.Labelframe(self.frm_lamp, text='Commands', padding=(PAD, PAD))
        self.lbf_lamp_comm.grid(row=0, column=0, sticky='nsew', padx=PAD, pady=PAD)
        self.lbf_lamp_comm.columnconfigure([0,1], weight=1)
        self.lbf_lamp_comm.rowconfigure([0,1,2], weight=1)

        self.lb_lamp_mode = ttk.Label(self.lbf_lamp_comm, text='Lamp Mode')
        self.lb_lamp_mode.grid(row=0, column=0)

        self.cbb_lamp_mode_on = ttk.Combobox(self.lbf_lamp_comm, width=7)
        self.cbb_lamp_mode_on.grid(row=0, column=1)
        self.cbb_lamp_mode_on.state(['readonly'])
        self.cbb_lamp_mode_on['values'] = ['On', 'Off']
        self.cbb_lamp_mode_on.set('Off')
        self.cbb_lamp_mode_on.bind('<<ComboboxSelected>>', self.lamp_mode)

        self.lb_lamp_debug = ttk.Label(self.lbf_lamp_comm, text='Lamp Debug')
        self.lb_lamp_debug.grid(row=1, column=0)

        self.cbb_lamp_debug_on = ttk.Combobox(self.lbf_lamp_comm, width=7)
        self.cbb_lamp_debug_on.grid(row=1, column=1)
        self.cbb_lamp_debug_on.state(['readonly'])
        self.cbb_lamp_debug_on['values'] = ['On', 'Off']
        self.cbb_lamp_debug_on.set('Off')
        self.cbb_lamp_debug_on.bind('<<ComboboxSelected>>', self.lamp_debug_on)

        self.bt_lamp_stats = ttk.Button(self.lbf_lamp_comm, text='Get Stats', width=12, command=self.lamp_stats)
        self.bt_lamp_stats.grid(row=2, column=0, columnspan=2)

        # Output
        self.lbf_lamp_out = ttk.Labelframe(self.frm_lamp, text='Output', padding=(PAD, PAD))
        self.lbf_lamp_out.grid(row=0, column=1, sticky='nsew', padx=(0, PAD), pady=PAD)
        self.lbf_lamp_out.columnconfigure([0], weight=1)
        self.lbf_lamp_out.rowconfigure([0], weight=1)

        self.txt_lamp = tk.Text(self.lbf_lamp_out, width=23, state='disabled', wrap='word', height=8)
        self.txt_lamp.grid(row=0, column=0, sticky='nsew')


    def widgets_state(self):
        # HEALTH
        if (not self.rng.health_startup_enabled) and (not self.rng.health_continuous_enabled):
            # Disable notebook tab
            tab_id = [tab_id for tab_id in self.nb.tabs() if 'health' in tab_id][0]
            self.nb.tab(tab_id, state=['disabled'])
        elif not self.rng.health_startup_enabled:
            # Disable health startup Labelframe
            self.lbf_health_startup.state(['disabled'])
            for child in self.lbf_health_startup.winfo_children():
                child.configure(state='disable')
        elif not self.rng.health_continuous_enabled:
            # Disable health startup Labelframe
            self.lbf_health_cont.state(['disabled'])
            for child in self.lbf_health_cont.winfo_children():
                child.configure(state='disable')

        # PERIPH
        if not self.rng.peripherals_enabled:
            # Disable notebook tab
            tab_id = [tab_id for tab_id in self.nb.tabs() if 'periph' in tab_id][0]
            self.nb.tab(tab_id, state=['disabled'])

        # LED
        if not self.led_enabled:
            # Disable notebook tab
            tab_id = [tab_id for tab_id in self.nb.tabs() if 'led' in tab_id][0]
            self.nb.tab(tab_id, state=['disabled'])

        # LAMP
        if not self.lamp_enabled:
            # Disable notebook tab
            tab_id = [tab_id for tab_id in self.nb.tabs() if 'lamp' in tab_id][0]
            self.nb.tab(tab_id, state=['disabled'])


    def notebook_tab_change(self, tk_event=None):
        # Retrieve info and update widgets
        if 'eeprom' in self.nb.select():
            self.eeprom_get()

        elif 'rng' in self.nb.select():
            self.rng_setup_get()
            self.rng_output_set('', '')

        elif 'health' in self.nb.select():
            self.health_startup_get()

        elif 'periph' in self.nb.select():
            self.periph_digi_state_update()

        elif 'led' in self.nb.select():
            self.led_color_scale()

        elif 'lamp' in self.nb.select():
            lamp_mode = 'On' if self.rng.lamp_mode else 'Off'
            lamp_debug = 'On' if self.rng.lamp_debug else 'Off'
            self.cbb_lamp_mode_on.set(lamp_mode)
            self.cbb_lamp_debug_on.set(lamp_debug)

        # Stop lamp
        if self.lamp_enabled and self.rng.lamp_mode:
            self.rng.snd_lamp_mode(False)

        # Stop bytes stream
        if self.rng.rng_streaming:
            self.rng_gen_byte_stream_stop()


    def eeprom_get(self):
        # Firmware
        firmware_dict = self.rng.get_eeprom_firmware()
        if firmware_dict:
            self.var_eeprom_firmw_version.set(firmware_dict['version'])
            modules_enabled = [key for key in firmware_dict if (key != 'version' and firmware_dict[key] == True)]
            modules_enabled = [m.replace('_enabled', '') for m in modules_enabled]
            self.var_eeprom_firmw_modules.set(modules_enabled)

        # PWM
        pwm = self.rng.get_eeprom_pwm()
        self.cbb_eeprom_pwm_freq.set(pwm['freq_str'])
        self.var_eeprom_pwm_duty.set(pwm['duty'])

        # RNG
        rng = self.rng.get_eeprom_rng()
        self.var_eeprom_rng_si.set(rng['sampling_interval_us'])

        # LED
        if self.rng.led_enabled:
            led = self.rng.get_eeprom_led()
            self.cbb_eeprom_led_attached.set('True' if led['led_attached'] else 'False')

        # LAMP
        if self.rng.led_enabled and self.rng.lamp_enabled:
            lamp = self.rng.get_eeprom_lamp()
            self.var_eeprom_lamp_exp_dur.set(lamp['exp_dur_max_ms'] // 1000)
            self.var_eeprom_lamp_z_sig.set('{:.3f}'.format(lamp['exp_z_significant']))
            self.var_eeprom_lamp_mag_smooth.set(lamp['exp_mag_smooth_n_trials'])


    def eeprom_set(self):
        if tkm.askyesno(message=('Save parameters to EEPROM?'), detail=''):
            # PWM
            freq_str = self.cbb_eeprom_pwm_freq.get()
            duty = self.var_eeprom_pwm_duty.get()
            self.rng.snd_eeprom_pwm(D_PWM_FREQ[freq_str], duty)

            # RNG
            sampling_interval_us = self.var_eeprom_rng_si.get()
            self.rng.snd_eeprom_rng(sampling_interval_us)

            # LED
            if self.rng.led_enabled:
                led_attached = True if self.cbb_eeprom_led_attached.get() == 'True' else False
                self.rng.snd_eeprom_led(led_attached)

            # LAMP
            if self.rng.led_enabled and self.rng.lamp_enabled:
                exp_dur_max_ms = self.var_eeprom_lamp_exp_dur.get() * 1000
                exp_z_significant = self.var_eeprom_lamp_z_sig.get()
                exp_mag_smooth_n_trials = self.var_eeprom_lamp_mag_smooth.get()
                self.rng.snd_eeprom_lamp(exp_dur_max_ms, exp_z_significant, exp_mag_smooth_n_trials)


    def eeprom_set_defaults(self):
        if tkm.askyesno(message=('Restore EEPROM to default values?'), detail=''):
            self.rng.snd_eeprom_reset_to_default()
            self.eeprom_get()


    def rng_notebook_tab_change(self, tk_event=None):
        # Clear data txts
        self.rng_output_set('', '')

        # Stop bytes stream
        if self.rng.rng_streaming:
            self.rng_gen_byte_stream_stop()


    def rng_setup_get(self):
        # PWM
        pwm = self.rng.get_pwm_setup()
        self.cbb_pwm_freq.set(pwm['freq_str'])
        self.var_pwm_duty.set(pwm['duty'])

        # RNG
        rng = self.rng.get_rng_setup()
        self.var_rng_si.set(rng['sampling_interval_us'])


    def rng_setup_set(self):
        # PWM
        freq_str = self.cbb_pwm_freq.get()
        duty = self.var_pwm_duty.get()
        self.rng.snd_pwm_setup(D_PWM_FREQ[freq_str], duty)

        # RNG
        sampling_interval_us = self.var_rng_si.get()
        self.rng.snd_rng_setup(sampling_interval_us)


    def rng_output_set(self, data_a, data_b):
        if data_a is None:
            data_a = ''
        data_a_fmt = '{}'.format(data_a).replace('[', '').replace(']', '').replace('\n', '')
        self.txt_rng_a.configure(state='normal')
        self.txt_rng_a.delete(1.0, 'end')
        self.txt_rng_a.insert(1.0, data_a_fmt)
        self.txt_rng_a.configure(state='disabled')

        if data_b is None:
            data_b = ''
        data_b_fmt = '{}'.format(data_b).replace('[', '').replace(']', '').replace('\n', '')
        self.txt_rng_b.configure(state='normal')
        self.txt_rng_b.delete(1.0, 'end')
        self.txt_rng_b.insert(1.0, data_b_fmt)
        self.txt_rng_b.configure(state='disabled')


    def rng_gen_pulse_count(self):
        n = self.var_rng_data_ammount.get()
        if n > CTRL_PANEL_N_MAX:
            n = CTRL_PANEL_N_MAX
            self.var_rng_data_ammount.set(CTRL_PANEL_N_MAX)

        pc_a, pc_b = self.rng.get_rng_pulse_counts(n, output_type='array')
        self.rng_output_set(pc_a, pc_b)


    def rng_gen_bit(self):
        bit_src = self.cbb_rng_data_bit_src.get()
        bit_src_id = D_RNG_BIT_SRC[bit_src]

        if bit_src == 'AB':
            bit_a, bit_b = self.rng.get_rng_bits(bit_src_id)
        elif bit_src == 'AB_RND':
            rng, bit = self.rng.get_rng_bits(bit_src_id)
            bit_a, bit_b = (bit, '') if rng == 'A' else ('', bit)
        elif bit_src == 'B':
            bit_a = ''
            bit_b = self.rng.get_rng_bits(bit_src_id)
        else:
            bit_a = self.rng.get_rng_bits(bit_src_id)
            bit_b = ''

        self.rng_output_set(bit_a, bit_b)


    def rng_gen_byte(self):
        n = self.var_rng_data_ammount.get()
        if n > CTRL_PANEL_N_MAX:
            n = CTRL_PANEL_N_MAX
            self.var_rng_data_ammount.set(CTRL_PANEL_N_MAX)

        postproc = self.cbb_rng_data_byte_pp.get()
        postproc_id = D_RNG_POSTPROC[postproc]

        bytes_a, bytes_b = self.rng.get_rng_bytes(n, postproc_id, output_type='array')        
        self.rng_output_set(bytes_a, bytes_b)


    def rng_gen_integer(self):
        n = self.var_rng_num_ammount.get()
        if n > CTRL_PANEL_N_MAX:
            n = CTRL_PANEL_N_MAX
            self.var_rng_num_ammount.set(CTRL_PANEL_N_MAX)
        min = self.var_rng_num_min_int.get()
        max = self.var_rng_num_max_int.get()

        delta = max - min
        if delta < 2**8:
            ints = self.rng.get_rng_int8s(n, delta, output_type='array').astype('u2')
        elif delta < 2**16:
            ints = self.rng.get_rng_int16s(n, delta, output_type='array').astype('u4')

        if ints is not None:
            ints += min
            self.rng_output_set(ints, '')


    def rng_gen_float(self):
        n = self.var_rng_num_ammount.get()
        if n > CTRL_PANEL_N_MAX:
            n = CTRL_PANEL_N_MAX
            self.var_rng_num_ammount.set(CTRL_PANEL_N_MAX)

        min = self.var_rng_num_min_float.get()
        max = self.var_rng_num_max_float.get()
        delta = max - min
        if delta < 0:
            self.lg.error('{} Floats: Provide delta >= 0'.format(self.name))
            return

        floats = self.rng.get_rng_floats(n, output_type='array')
        if floats is not None:
            floats = min + delta * floats            
            self.rng_output_set(floats, '')


    def rng_gen_byte_stream_start(self):
        n = self.var_rng_stream_ammount.get()
        if n > CTRL_PANEL_N_MAX:
            n = CTRL_PANEL_N_MAX
            self.var_rng_stream_ammount.set(CTRL_PANEL_N_MAX)

        interval_ms = self.var_rng_stream_interval.get()
        if interval_ms > RNG_BYTE_STREAM_MAX_INTERVAL_MS:
            interval_ms = RNG_BYTE_STREAM_MAX_INTERVAL_MS
            self.var_rng_stream_interval.set(RNG_BYTE_STREAM_MAX_INTERVAL_MS)

        postproc = self.cbb_rng_stream_byte_pp.get()
        postproc_id = D_RNG_POSTPROC[postproc]

        # Register callback
        self.rng.cbkreg_stream_data_available(self.rng_byte_stream_data)

        # Start stream
        self.rng.snd_rng_byte_stream_start(n, interval_ms, postproc_id)


    def rng_gen_byte_stream_stop(self, tk_event=None):
        if self.rng.rng_streaming:
            self.rng.snd_rng_byte_stream_stop()


    def rng_byte_stream_data(self):
        bytes_a, bytes_b = self.rng.get_rng_byte_stream_data(output_type='array')
        self.rng_output_set(bytes_a, bytes_b)


    def health_results_set(self, results_str):
        self.txt_health_results.configure(state='normal')
        self.txt_health_results.delete(1.0, 'end')
        self.txt_health_results.insert(1.0, results_str)
        self.txt_health_results.configure(state='disabled')


    def health_startup_run(self):
        self.rng.snd_health_startup_run()
        self.after(750, self.health_startup_get)


    def health_startup_get(self):
        success, res = self.rng.get_health_startup_results()
        res_str = 'Pulse Count = {}\n  RNG_A = {:.1f}  RNG_B = {:.1f}  TRESH = {:.1f}\n\n'\
                  'Pulse Count Difference = {}\n  RNG_A = {:.1f}  RNG_B = {:.1f}  TRESH = {:.1f}\n\n'\
                  'Bit Bias = {}\n  RNG_A = {:.2f}  RNG_B = {:.2f}  TRESH = {:.2f}\n\n'\
                  'Byte Bias = {}\n  RNG_A = {:.1f}  RNG_B = {:.1f}  TRESH = {:.1f}\n\n'\
                  'Startup Result = {}'\
                  .format(*res['pulse_count'], *res['pulse_count_diff'], *res['bit_bias'], *res['byte_bias'], success)

        self.health_results_set(res_str)


    def health_continuous_get(self):
        errors_n, res = self.rng.get_health_continuous_errors()
        res_str = 'Repetitive Count Errors\n  RNG_A = {}\n  RNG_B = {}\n\n'\
                  'Adaptative Proportion Errors\n  RNG_A = {}\n  RNG_B = {}\n\n'\
                  'Total Errors = {}'\
                  .format(res['repetitive_count_a'], res['repetitive_count_b'],
                          res['adaptative_proportion_a'], res['adaptative_proportion_b'], errors_n)

        self.health_results_set(res_str)


    def periph_digi_state_update(self, tk_event=None):
        # Update mode and state
        port_id = self.var_periph_port.get()
        state, mode = self.rng.get_periph_digi_state(port_id)
        self.cbb_periph_mode.set(mode)
        self.var_periph_state.set(state)


    def periph_digi_mode(self, tk_event=None):
        port_id = self.var_periph_port.get()
        mode = self.cbb_periph_mode.get()
        mode_id = 0 if mode == 'INPUT' else 1

        self.rng.snd_periph_digi_mode(port_id, mode_id)


    def periph_digi_state_get(self):
        port_id = self.var_periph_port.get()

        # Change mode?
        if self.cbb_periph_mode.get() == 'OUTPUT':
            self.cbb_periph_mode.set('INPUT')
            self.periph_digi_mode()

        state, mode = self.rng.get_periph_digi_state(port_id)
        if state is not None:
            self.var_periph_state.set(state)


    def periph_digi_state_set(self):
        port_id = self.var_periph_port.get()

        # Change mode?
        if self.cbb_periph_mode.get() == 'INPUT':
            self.cbb_periph_mode.set('OUTPUT')
            self.periph_digi_mode()

        state = self.var_periph_state.get()
        self.rng.snd_periph_digi_state(port_id, state)


    def periph_digi_pulse(self):
        port_id = self.var_periph_port.get()

        pulse_dur_us = self.var_periph_pulse_dur.get()
        if pulse_dur_us > 100000:
            pulse_dur_us = 100000
            self.var_periph_pulse_dur.set(100000)

        # Change mode?
        if self.cbb_periph_mode.get() == 'INPUT':
            self.cbb_periph_mode.set('OUTPUT')
            self.periph_digi_mode()

        self.rng.snd_periph_digi_pulse(port_id, pulse_dur_us)


    def periph_d1_timing_debug(self, tk_event=None):
        on = True if self.cbb_periph_d1_timing_dbg.get() == 'On' else False
        self.rng.snd_rng_timing_debug_d1(on)


    def periph_d1_trigger_input(self, tk_event=None):
        on = True if self.cbb_periph_d1_trigger_on.get() == 'On' else False
        self.rng.snd_periph_d1_trigger_input(on)


    def periph_d1_comparator(self, tk_event=None):
        neg = True if self.cbb_periph_d1_comparator_neg.get() == 'Neg=D5' else False
        on = True if self.cbb_periph_d1_comparator_on.get() == 'On' else False
        self.rng.snd_periph_d1_comparator(neg, on)


    def periph_d1_delay_test(self, tk_event=None):
        delay_us = self.var_periph_d1_delay_dur.get()
        if delay_us > 100000:
            delay_us = 100000
            self.var_periph_d1_delay_dur.set(100000)

        self.rng.snd_periph_d1_delay_us_test(delay_us)


    def periph_d2_input_capture(self, tk_event=None):
        on = True if self.cbb_periph_d2_input_capture_on.get() == 'On' else False
        self.rng.snd_periph_d2_input_capture(on)

        # Register callback
        if on:
            self.rng.cbkreg_d2_input_capture_available(self.periph_d2_input_capture_data)
        else:
            self.rng.cbkreg_d2_input_capture_available(None)


    def periph_d2_input_capture_data(self):
        interval_s = self.rng.get_periph_d2_input_capture()
        self.var_periph_d2_input_capture_interval.set('{:.3f}'.format(interval_s))


    def periph_d3_trigger_output(self, tk_event=None):
        on = True if self.cbb_periph_d3_trigger_out_on.get() == 'On' else False

        interval_ms = self.var_periph_d3_trigger_out_interval.get()
        if interval_ms is None:
            return
        if interval_ms > RNG_BYTE_STREAM_MAX_INTERVAL_MS:
            interval_ms = RNG_BYTE_STREAM_MAX_INTERVAL_MS
            self.var_periph_d3_trigger_out_interval.set(RNG_BYTE_STREAM_MAX_INTERVAL_MS)

        self.rng.snd_periph_d3_timer3_trigger_output(interval_ms, on)


    def periph_d3_pwm(self, tk_event=None):
        on = True if self.cbb_periph_d3_pwm_on.get() == 'On' else False

        freq_prescaler = self.var_periph_d3_pwm_presc.get()
        top = self.var_periph_d3_pwm_top.get()
        duty = int(self.var_periph_d3_pwm_duty.get()/100 * 65535)

        self.rng.snd_periph_d3_timer3_pwm(freq_prescaler, top, duty, on)


    def periph_d4_pin_change(self, tk_event=None):
        on = True if self.cbb_periph_d4_pin_change_on.get() == 'On' else False
        self.rng.snd_periph_d4_pin_change(on)


    def periph_d5_adc_read(self):
        ref_5v = True if self.cbb_periph_d5_adc_ref.get() == '5V' else False
        clk_prescaler = self.var_periph_d5_adc_prescaler.get()
        oversampling_n_bits = self.var_periph_d5_adc_oversampling.get()

        adc_read = self.rng.get_periph_d5_adc_read(ref_5v, clk_prescaler, oversampling_n_bits, on=True)
        self.var_periph_d5_adc.set('{:.3f}'.format(adc_read))


    def led_color_scale(self, tk_event=None):
        color_hue = self.var_led_color.get()
        intensity = self.var_led_int_scale.get()

        color_str = ''
        if color_hue in D_LED_COLOR_INV:
            color_str = D_LED_COLOR_INV[color_hue]
        self.var_led_color_str.set(color_str)
        self.rng.snd_led_color(color_hue, intensity)


    def led_color_cbb(self, tk_event=None):
        color_str = self.var_led_color_str.get()
        color_hue = D_LED_COLOR[color_str]
        intensity = self.var_led_int_scale.get()

        self.var_led_color.set(color_hue)
        self.rng.snd_led_color(color_hue, intensity)


    def led_color_fade(self, tk_event=None):
        color_str = self.var_led_color_fade_str.get()
        color_hue = D_LED_COLOR[color_str]
        duration_ms = self.var_led_color_fade_dur.get()

        if self.rng.snd_led_color_fade(color_hue, duration_ms):
            self.after(duration_ms, lambda: [self.var_led_color.set(color_hue), self.var_led_color_str.set(color_str)])


    def led_color_oscillate(self):
        n_cycles = self.var_led_color_osc_cycles.get()
        duration_ms = self.var_led_color_osc_dur.get()
        self.rng.snd_led_color_oscillate(n_cycles, duration_ms)


    def led_intensity_scale(self, tk_event=None):
        intensity = self.var_led_int_scale.get()

        self.var_led_int_spb.set(intensity)
        self.rng.snd_led_intensity(intensity)


    def led_intensity_spb(self, tk_event=None):
        intensity = self.var_led_int_spb.get()

        self.var_led_int_scale.set(intensity)
        self.rng.snd_led_intensity(intensity)


    def led_intensity_fade(self, tk_event=None):
        intensity = self.var_led_int_fade.get()
        duration_ms = self.var_led_int_fade_dur.get()

        if self.rng.snd_led_intensity_fade(intensity, duration_ms):
            self.after(duration_ms, lambda: [self.var_led_int_scale.set(intensity), self.var_led_int_spb.set(intensity)])


    def lamp_mode(self, tk_event=None):
        on = True if self.cbb_lamp_mode_on.get() == 'On' else False

        self.rng.snd_lamp_mode(on)


    def lamp_debug_on(self, tk_event=None):
        on = True if self.cbb_lamp_debug_on.get() == 'On' else False

        self.rng.snd_lamp_debug(on)

        # Register callback
        if on:
            self.rng.cbkreg_lamp_debug(self.lamp_debug_data)
        else:
            self.rng.cbkreg_lamp_debug(None)


    def lamp_output_set(self, output_str):
        self.txt_lamp.configure(state='normal')
        self.txt_lamp.delete(1.0, 'end')
        self.txt_lamp.insert(1.0, output_str)
        self.txt_lamp.configure(state='disabled')


    def lamp_debug_data(self):
        dbg_dict = self.rng.get_lamp_debug_data()

        debug_str = 'LAMP Debug\n\nz={:+1.2f}, mag={:3.0f}, mag_avg={:3.0f}'\
                    .format(dbg_dict['z'], dbg_dict['mag'], dbg_dict['mag_avg'])
        self.lamp_output_set(debug_str)


    def lamp_stats(self, tk_event=None):
        stats = self.rng.get_lamp_statistics()
        exp_n = stats['exp_n']

        if exp_n:
            res_str = 'LAMP Statistics\n\n'\
                      'Total of {} experiments, where {} ({:.2f}%) reached statistical significance\n\n'\
                      'Color distribution (%):\n'\
                      '  R={:.2f} O={:.2f} Y={:.2f} G={:.2f}\n'\
                      '  C={:.2f} B={:.2f} PU={:.2f} PI={:.2f}\n'\
                      .format(exp_n, stats['exp_n_zsig'], stats['exp_n_zsig']/exp_n*100,
                              stats['red']/exp_n*100, stats['orange']/exp_n*100, stats['yellow']/exp_n*100,
                              stats['green']/exp_n*100, stats['cyan']/exp_n*100, stats['blue']/exp_n*100,
                              stats['purple']/exp_n*100, stats['pink']/exp_n*100)

        else:
            res_str = 'LAMP Statistics\n\n'\
                      'No experiment recorded'

        self.lamp_output_set(res_str)


rava_subapp_control_panel = {'class':RAVA_SUBAPP_CTRL_PANEL,
                             'menu_title':'Control Panel',
                             'show_button':True,
                             'use_rng':True
                             }


"""
Copyright (c) 2024 Gabriel Guerrer

Distributed under the MIT license - See LICENSE for details
"""

"""
The Acquisition sub-app features a GUI interface for utilizing the 
RAVA_ACQUISITION class, which allows the generation of files comprising pulse 
counts, bytes, or numbers extracted from a RAVA device.
"""

from glob import glob
import os
import os.path
import time
import datetime

import tkinter as tk
from tkinter import ttk
import tkinter.messagebox as tkm

import numpy as np

from rng_rava.acq import RAVA_ACQUISITION, ACQ_BYTES_RNG_OUT, get_ammount_prefix_number, get_ammount_prefix_str
from rng_rava.rava_defs import D_RNG_POSTPROC
from rng_rava.tk import RAVA_SUBAPP
from rng_rava.tk.acq import WIN_PROGRESS

### VARS

PAD = 10


### RAVA_SUBAPP_ACQUISITION

class RAVA_SUBAPP_ACQUISITION(RAVA_SUBAPP):

    cfg_default_str = '''
    [ACQUISITION]        
    chunk_bytes = 10000
    chunk_numbers = 5000
    delete_incomplete = True
    path_out = 
    '''

    def __init__(self, parent):
        ## Initialize RAVA_SUBAPP
        name = 'RAVA_SUBAPP ACQUISITION'
        win_title = 'Acquisition'
        win_geometry = '510x280'
        win_resizable = False
        if not super().__init__(parent, name=name, win_title=win_title, win_geometry=win_geometry, win_resizable=win_resizable):
            return

        # Windows
        self.win_progress = WIN_PROGRESS(self)
        self.win_progress.hide()        

        # Widgets
        self.nb = ttk.Notebook(self, padding=PAD)        
        self.nb.grid(row=0, column=0, sticky='nsew')

        self.frm_pcs = ttk.Frame(self, name='pulsecounts')
        self.frm_pcs.grid(row=0, column=0, sticky='nsew')
        self.pcs_widgets()

        self.frm_bytes = ttk.Frame(self, name='bytes')
        self.frm_bytes.grid(row=0, column=0, sticky='nsew')
        self.bytes_widgets()

        self.nfrm_nums = ttk.Frame(self, name='numbers')
        self.nfrm_nums.grid(row=0, column=0, sticky='nsew')
        self.nums_widgets()

        self.nb.add(self.frm_pcs, text=' Pulse Counts ')
        self.nb.add(self.frm_bytes, text=' Bytes ')
        self.nb.add(self.nfrm_nums, text=' Numbers ')
        
        ## Start
        
        # Config        
        delete_incomplete = self.cfg.read('ACQUISITION', 'delete_incomplete', bool)
        out_path = self.cfg.read('ACQUISITION', 'path_out')
        self.var_acq_path.set(out_path)

        # RAVA_ACQUISITION
        self.rng_acq = RAVA_ACQUISITION(delete_incomplete=delete_incomplete)
        self.rng_acq.cbkreg_progress(self.win_progress.prog_update)        
        

    def pcs_widgets(self):
        self.frm_pcs.columnconfigure([0], weight=1)
        self.frm_pcs.columnconfigure([1], weight=7)
        self.frm_pcs.rowconfigure([0,1,2], weight=1)

        ## Path
        self.lb_pcs_path = ttk.Label(self.frm_pcs, text='Out Path')
        self.lb_pcs_path.grid(row=0, column=0)

        self.var_acq_path = tk.StringVar(value='')
        self.en_pcs_path = ttk.Entry(self.frm_pcs, textvariable=self.var_acq_path)
        self.en_pcs_path.grid(row=0, column=1, sticky='ew')

        self.bt_pcs_path_search = ttk.Button(self.frm_pcs, width=2, text='...', command=self.acq_path_search)
        self.bt_pcs_path_search.grid(row=0, column=2, padx=PAD)

        ## PCs parameters
        self.frm_pcs_pars = ttk.Frame(self.frm_pcs, name='pcs_pars')
        self.frm_pcs_pars.grid(row=1, column=0, columnspan=3, sticky='nsew', padx=2*PAD)
        self.frm_pcs_pars.columnconfigure([0,1,2,3], weight=1)
        self.frm_pcs_pars.rowconfigure([0,1,2], weight=1)

        # Sampling Interval
        self.lb_pcs_si = ttk.Label(self.frm_pcs_pars, text='SI (min, max, step)')
        self.lb_pcs_si.grid(row=0, column=0, sticky='w')

        self.var_pcs_si_min = tk.IntVar(value=10)
        self.spb_pcs_si_min = ttk.Spinbox(self.frm_pcs_pars, from_=1, to=100, increment=1, textvariable=self.var_pcs_si_min, width=8)
        self.spb_pcs_si_min.grid(row=0, column=1, sticky='w')

        self.var_pcs_si_max = tk.IntVar(value=10)
        self.spb_pcs_si_max = ttk.Spinbox(self.frm_pcs_pars, from_=1, to=100, increment=1, textvariable=self.var_pcs_si_max, width=8)
        self.spb_pcs_si_max.grid(row=0, column=2, sticky='w')

        self.var_pcs_si_step = tk.IntVar(value=1)
        self.spb_pcs_si_step = ttk.Spinbox(self.frm_pcs_pars, from_=1, to=100, increment=1, textvariable=self.var_pcs_si_step, width=8)
        self.spb_pcs_si_step.grid(row=0, column=3, sticky='w')

        # N PCs
        self.lb_pcs_n = ttk.Label(self.frm_pcs_pars, text='N PCs per SI')
        self.lb_pcs_n.grid(row=1, column=0, sticky='w')

        self.var_pcs_n = tk.DoubleVar(value=50)
        self.spb_pcs_n = ttk.Spinbox(self.frm_pcs_pars, from_=1, to=999, increment=1, textvariable=self.var_pcs_n, width=8)
        self.spb_pcs_n.grid(row=1, column=1, sticky='w')

        self.cbb_pcs_n_mult = ttk.Combobox(self.frm_pcs_pars, width=8)
        self.cbb_pcs_n_mult.grid(row=1, column=2, sticky='w')
        self.cbb_pcs_n_mult.state(['readonly'])
        self.cbb_pcs_n_mult['values'] = ['', 'K', 'M', 'G', 'T']
        self.cbb_pcs_n_mult.set('K')

        # Preset
        self.lb_pcs_preset = ttk.Label(self.frm_pcs_pars, text='Preset')
        self.lb_pcs_preset.grid(row=2, column=0, sticky='w')
        
        self.cbb_pcs_preset = ttk.Combobox(self.frm_pcs_pars, width=8, )
        self.cbb_pcs_preset.grid(row=2, column=1, sticky='w')
        self.cbb_pcs_preset.state(['readonly'])
        self.cbb_pcs_preset['values'] = ['Report', 'Detailed']        
        self.cbb_pcs_preset.bind('<<ComboboxSelected>>', self.pcs_preset_sel)
        self.cbb_pcs_preset.set('Report')

        ## Generate
        self.bt_pcs_generate = ttk.Button(self.frm_pcs, text='Generate', command=self.pcs_generate)
        self.bt_pcs_generate.grid(row=2, column=0, columnspan=3)


    def bytes_widgets(self):
        self.frm_bytes.columnconfigure([0], weight=1)
        self.frm_bytes.columnconfigure([1], weight=7)
        self.frm_bytes.rowconfigure([0,1,2], weight=1)

        ## Path
        self.lb_bytes_path = ttk.Label(self.frm_bytes, text='Out Path')
        self.lb_bytes_path.grid(row=0, column=0)

        self.en_bytes_path = ttk.Entry(self.frm_bytes, textvariable=self.var_acq_path)
        self.en_bytes_path.grid(row=0, column=1, sticky='ew')

        self.bt_bytes_path_search = ttk.Button(self.frm_bytes, width=2, text='...', command=self.acq_path_search)
        self.bt_bytes_path_search.grid(row=0, column=2, padx=PAD)

        ## Bin pars
        self.frm_bytes_pars = ttk.Frame(self.frm_bytes, name='binary_pars')
        self.frm_bytes_pars.grid(row=1, column=0, columnspan=3, sticky='nsew', padx=2*PAD)
        self.frm_bytes_pars.columnconfigure([0,1,2,3], weight=1)
        self.frm_bytes_pars.rowconfigure([0,1,2], weight=1)

        # N bytes
        self.lb_bytes_n = ttk.Label(self.frm_bytes_pars, text='N Bytes')
        self.lb_bytes_n.grid(row=0, column=0, sticky='w')

        self.var_bytes_n = tk.DoubleVar(value=125)
        self.spb_bytes_n = ttk.Spinbox(self.frm_bytes_pars, from_=1, to=999, increment=1, textvariable=self.var_bytes_n, width=12)
        self.spb_bytes_n.grid(row=0, column=1, sticky='w')

        self.cbb_bytes_n_mult = ttk.Combobox(self.frm_bytes_pars, width=8)
        self.cbb_bytes_n_mult.grid(row=0, column=2, sticky='w')
        self.cbb_bytes_n_mult.state(['readonly'])
        self.cbb_bytes_n_mult['values'] = ['B', 'KB', 'MB', 'GB', 'TB']
        self.cbb_bytes_n_mult.set('MB')

        # Post-proc
        self.lb_bytes_postproc = ttk.Label(self.frm_bytes_pars, text='Post-proc')
        self.lb_bytes_postproc.grid(row=1, column=0, sticky='w')

        self.cbb_bytes_postproc = ttk.Combobox(self.frm_bytes_pars, width=13)
        self.cbb_bytes_postproc.grid(row=1, column=1, sticky='w')
        self.cbb_bytes_postproc.state(['readonly'])
        self.cbb_bytes_postproc['values'] = list(D_RNG_POSTPROC.keys())
        self.cbb_bytes_postproc.set('NONE')

        # Output rng cores
        self.lb_bytes_rngout = ttk.Label(self.frm_bytes_pars, text='RNG Out')
        self.lb_bytes_rngout.grid(row=1, column=2, sticky='w')

        self.cbb_bytes_rngout = ttk.Combobox(self.frm_bytes_pars, width=10)
        self.cbb_bytes_rngout.grid(row=1, column=3, sticky='w')
        self.cbb_bytes_rngout.state(['readonly'])
        self.cbb_bytes_rngout['values'] = ACQ_BYTES_RNG_OUT
        self.cbb_bytes_rngout.set('AB')

        # Preset
        self.lb_bytes_preset = ttk.Label(self.frm_bytes_pars, text='Preset')
        self.lb_bytes_preset.grid(row=2, column=0, sticky='w')
        
        self.cbb_bytes_preset = ttk.Combobox(self.frm_bytes_pars, width=8, )
        self.cbb_bytes_preset.grid(row=2, column=1, sticky='w')
        self.cbb_bytes_preset.state(['readonly'])
        self.cbb_bytes_preset['values'] = ['Report']        
        self.cbb_bytes_preset.bind('<<ComboboxSelected>>', self.bytes_preset_sel)
        self.cbb_bytes_preset.set('Report')

        ## Generate
        self.bt_bytes_generate = ttk.Button(self.frm_bytes, text='Generate', command=self.bytes_generate)
        self.bt_bytes_generate.grid(row=2, column=0, columnspan=3)


    def nums_widgets(self):
        self.nfrm_nums.columnconfigure([0], weight=1)
        self.nfrm_nums.columnconfigure([1], weight=7)
        self.nfrm_nums.rowconfigure([0,1,2], weight=1)

        ## Path
        self.lb_nums_path = ttk.Label(self.nfrm_nums, text='Out Path')
        self.lb_nums_path.grid(row=0, column=0)

        self.en_nums_path = ttk.Entry(self.nfrm_nums, textvariable=self.var_acq_path)
        self.en_nums_path.grid(row=0, column=1, sticky='ew')

        self.bt_nums_path_search = ttk.Button(self.nfrm_nums, width=2, text='...', command=self.acq_path_search)
        self.bt_nums_path_search.grid(row=0, column=2, padx=PAD)

        ## Number options
        self.nfrm_nums_pars = ttk.Frame(self.nfrm_nums, name='number_pars')
        self.nfrm_nums_pars.grid(row=1, column=0, columnspan=3, sticky='nsew', padx=2*PAD)
        self.nfrm_nums_pars.columnconfigure([0,1,2,3], weight=1)
        self.nfrm_nums_pars.rowconfigure([0,1,2], weight=1)

        # Number type
        self.lb_nums_type = ttk.Label(self.nfrm_nums_pars, text='Number (type, N)')
        self.lb_nums_type.grid(row=0, column=0, sticky='w')
        
        self.cbb_nums_type = ttk.Combobox(self.nfrm_nums_pars, width=9)
        self.cbb_nums_type.grid(row=0, column=1, sticky='w')
        self.cbb_nums_type.state(['readonly'])
        self.cbb_nums_type['values'] = ['INT', 'FLOAT']
        self.cbb_nums_type.set('INT')
        self.cbb_nums_type.bind('<<ComboboxSelected>>', self.nums_type_sel)

        self.var_nums_n = tk.DoubleVar(value=10)
        self.spb_nums_n = ttk.Spinbox(self.nfrm_nums_pars, from_=1, to=999, increment=1, textvariable=self.var_nums_n, width=8)
        self.spb_nums_n.grid(row=0, column=2, sticky='w')

        self.cbb_nums_n_mult = ttk.Combobox(self.nfrm_nums_pars, width=9)
        self.cbb_nums_n_mult.grid(row=0, column=3, sticky='w')
        self.cbb_nums_n_mult.state(['readonly'])
        self.cbb_nums_n_mult['values'] = ['', 'K', 'M', 'G', 'T']
        self.cbb_nums_n_mult.set('M')        

        # Number range
        self.lb_nums_range = ttk.Label(self.nfrm_nums_pars, text='Range (min, max)')
        self.lb_nums_range.grid(row=1, column=0, sticky='w')

        self.var_nums_range_min = tk.DoubleVar(value=0)
        self.spb_nums_range_min = ttk.Spinbox(self.nfrm_nums_pars, from_=1, to=999, increment=1, textvariable=self.var_nums_range_min, width=8)
        self.spb_nums_range_min.grid(row=1, column=1, sticky='w')

        self.var_nums_range_max = tk.DoubleVar(value=99)
        self.spb_nums_range_max = ttk.Spinbox(self.nfrm_nums_pars, from_=1, to=999, increment=1, textvariable=self.var_nums_range_max, width=8)
        self.spb_nums_range_max.grid(row=1, column=2, sticky='w')

        # Binary output
        self.lb_nums_out_type = ttk.Label(self.nfrm_nums_pars, text='Out Type')
        self.lb_nums_out_type.grid(row=2, column=0, sticky='w')

        self.cbb_nums_out_type = ttk.Combobox(self.nfrm_nums_pars, width=9)
        self.cbb_nums_out_type.grid(row=2, column=1, sticky='w')
        self.cbb_nums_out_type.state(['readonly'])
        self.cbb_nums_out_type['values'] = ['Binary', 'Text']
        self.cbb_nums_out_type.set('Binary')
        self.cbb_nums_out_type.bind('<<ComboboxSelected>>', self.nums_out_type_sel)

        # Separator
        self.lb_nums_out_sep = ttk.Label(self.nfrm_nums_pars, text='Text Separator')
        self.lb_nums_out_sep.grid(row=2, column=2, sticky='w')

        self.cbb_nums_out_sep = ttk.Combobox(self.nfrm_nums_pars, width=9)
        self.cbb_nums_out_sep.grid(row=2, column=3, sticky='w')
        self.cbb_nums_out_sep.state(['readonly'])
        self.cbb_nums_out_sep['values'] = ['LINE', ',', ';']
        self.cbb_nums_out_sep.set('LINE')
        self.cbb_nums_out_sep.state(['disabled'])

        ## Generate
        self.bt_nums_generate = ttk.Button(self.nfrm_nums, text='Generate', command=self.nums_generate)
        self.bt_nums_generate.grid(row=2, column=0, columnspan=3)


    def acquisition_finished(self, future):
        # Get result
        outputs, progress, time_str = future.result()

        if progress == 0:
            tkm.showerror(parent=self, title='Error', message='Acquisition failed')
            return

        # Show result info
        if progress == 100:
            title_msg = 'Success'
        else: 
            title_msg = 'Canceled'

        time_msg = 'Time = ' + time_str
        
        if len(outputs):
            detail_msg = '\nGenerated files\n{}'.format(outputs)
        else: 
            detail_msg = '\nIncomplete files deleted'

        tkm.showinfo(parent=self.win_progress, title=title_msg, message=time_msg, detail=detail_msg)        

        # Close win_progress
        self.win_progress.hide()


    def acq_path_search(self):
        out_path0 = self.var_acq_path.get()        
        out_path0 = out_path0 if out_path0 and os.path.exists(out_path0) else '/'
        out_path = tk.filedialog.askdirectory(parent=self, initialdir=out_path0, mustexist=True)
        if out_path:
            self.var_acq_path.set(out_path)
            self.cfg.write('ACQUISITION', 'path_out', out_path)

    
    def pcs_preset_sel(self, tk_event=None):
        if self.cbb_pcs_preset.get() == 'Report':
            self.var_pcs_si_min.set(10)
            self.var_pcs_si_max.set(10)
            self.var_pcs_si_step.set(1)
            self.var_pcs_n.set(50)
            self.cbb_pcs_n_mult.set('K')

        elif self.cbb_pcs_preset.get() == 'Detailed':
            self.var_pcs_si_min.set(1)
            self.var_pcs_si_max.set(10)
            self.var_pcs_si_step.set(1)
            self.var_pcs_n.set(10)
            self.cbb_pcs_n_mult.set('M')

    
    def pcs_generate(self):
        # Read vars
        out_path = self.var_acq_path.get()
        if not os.access(out_path, os.W_OK):
            tkm.showerror(parent=self, message='Error', detail='Cannot write on the chosen directory')
            return
        
        si_min = self.var_pcs_si_min.get()
        si_max = self.var_pcs_si_max.get()
        si_step = self.var_pcs_si_step.get()        
        si_range = np.arange(si_min, si_max + 1, si_step)
        n_si_range = len(si_range)
        if n_si_range == 0:
            tkm.showerror(parent=self, message='Error', detail='Please choose a proper Sampling Interval range')
            return

        n_pcs = int(get_ammount_prefix_number(n=self.var_pcs_n.get(), prefix=self.cbb_pcs_n_mult.get()))
        n_chunk = self.cfg.read('ACQUISITION', 'chunk_bytes', int)

        # Save cfg
        self.cfg.write('ACQUISITION', 'path_out', out_path)

        # Info
        self.lg.info('{}: Generating {} Pulse Counts in the range ({}, {}, {})'
                     .format(self.name, n_pcs, n_si_range, si_min, si_max, si_step))

        # Start
        time_start = time.perf_counter()

        # Data
        pcs_a = np.zeros(shape=(n_si_range, n_pcs), dtype=np.uint8)
        pcs_b = np.zeros(shape=(n_si_range, n_pcs), dtype=np.uint8)

        # Save initial RAVA setup
        pwm_setup = self.rng.get_pwm_setup()
        rng_setup = self.rng.get_rng_setup()

        # Show progress window
        with self.win_progress:

            # Loop PCs acquisition        
            canceled = False
            for i, si in enumerate(si_range):

                # Update progress window title
                self.win_progress.set_extra_title('{}/{}'.format(i+1, n_si_range))

                # Set Sampling Interval
                self.rng.snd_rng_setup(si)

                # Collect PCs
                outputs, _, _ = self.rng_acq.get_pulse_counts(rng=self.rng, n_pcs=n_pcs, n_chunk=n_chunk, rng_out='AB', 
                                                            out_file=False, out_path='', threaded=False)
                if len(outputs):
                    pcs_a[i], pcs_b[i] = outputs

                # Canceled?
                if not self.win_progress.acquiring:
                    canceled = True
                    break

        # Restore initial SI
        self.rng.snd_rng_setup(**rng_setup)

        # Timing
        time_s = time.perf_counter() - time_start
        time_td = datetime.timedelta(seconds=time_s)
        time_str = '{}'.format(str(time_td).split('.')[0])

        # Finished without cancelling?
        if not canceled:
            # Save Data
            sn = self.rng.dev_serial_number
            n_pcs_str = get_ammount_prefix_str(n_pcs)
            out_name = '{}_PCS_{}_SI_{}_{}_{}__'.format(sn, n_pcs_str, si_min, si_max, si_step)
            n_out_name = len(glob(os.path.join(out_path, out_name + '*.npz')))
            output_file = os.path.join(out_path, out_name + '{}.npz'.format(n_out_name + 1))
            
            np.savez_compressed(output_file, pwm_setup=pwm_setup, rng_setup=rng_setup, si_range=si_range, pcs_a=pcs_a, pcs_b=pcs_b)

            # Info screen
            tkm.showinfo(parent=self, title='Success', message='Time = ' + time_str, detail='\nGenerated file\n{}'.format([output_file]))

        else:
            # Info screen
            tkm.showinfo(parent=self, title='Canceled', message='Time = ' + time_str, detail='\nIncomplete files deleted')


    def bytes_preset_sel(self, tk_event=None):
        if self.cbb_bytes_preset.get() == 'Report':
            self.var_bytes_n.set(125)
            self.cbb_bytes_n_mult.set('MB')
            self.cbb_bytes_postproc.set('NONE')
            self.cbb_bytes_rngout.set('AB')


    def bytes_generate(self):
        # Read vars
        out_path = self.var_acq_path.get()
        if not os.access(out_path, os.W_OK):
            tkm.showerror(parent=self, message='Error', detail='Cannot write on the chosen directory')
            return

        n_bytes = int(get_ammount_prefix_number(n=self.var_bytes_n.get(), prefix=self.cbb_bytes_n_mult.get()))
        postproc = self.cbb_bytes_postproc.get()
        rng_out = self.cbb_bytes_rngout.get()
        n_chunk = self.cfg.read('ACQUISITION', 'chunk_bytes', int)

        # Save cfg
        self.cfg.write('ACQUISITION', 'path_out', out_path)

        # Info
        self.lg.info('{}: Generating {} Bytes; postproc={}, rng_out={}'.format(self.name, n_bytes, postproc, rng_out))

        # Start byte acquisition
        self.win_progress.show()        

        result_future = self.rng_acq.get_bytes(rng=self.rng, n_bytes=n_bytes, n_chunk=n_chunk, postproc=postproc, 
                                               rng_out=rng_out, out_file=True, out_path=out_path, threaded=True)
        result_future.add_done_callback(self.acquisition_finished)


    def nums_type_sel(self, tk_event=None):
        if self.cbb_nums_type.get() == 'INT':
            self.var_nums_range_min.set(0)
            self.var_nums_range_max.set(999)
        else:
            self.var_nums_range_min.set(0)
            self.var_nums_range_max.set(1)


    def nums_out_type_sel(self, tk_event=None):
        if self.cbb_nums_out_type.get() == 'Binary':
            self.cbb_nums_out_sep.state(['disabled'])
        else:
            self.cbb_nums_out_sep.state(['!disabled'])


    def nums_generate(self):
        ## Read vars
        out_path = self.var_acq_path.get()
        if not os.access(out_path, os.W_OK):
            tkm.showerror(parent=self, title='Error', message='Cannot write on the chosen directory')
            return
        
        num_type_str = self.cbb_nums_type.get()
        num_type = int if num_type_str == 'INT' else float
        n_nums = int(get_ammount_prefix_number(n=self.var_nums_n.get(), prefix=self.cbb_nums_n_mult.get()))
        num_min = self.var_nums_range_min.get()
        num_max = self.var_nums_range_max.get()
        out_binary = True if self.cbb_nums_out_type.get() == 'Binary' else False
        out_separator = self.cbb_nums_out_sep.get()
        out_separator = '\n' if out_separator == 'LINE' else out_separator
        n_chunk = self.cfg.read('ACQUISITION', 'chunk_numbers', int)

        # Save cfg
        self.cfg.write('ACQUISITION', 'path_out', out_path)

        # Info
        self.lg.info('{}: Generating {} {} Numbers; range=[{}, {}]'
                     .format(self.name, n_nums, num_type_str, num_min, num_max))

        # Start number acquisition
        self.win_progress.show()        

        result_future = self.rng_acq.get_numbers(rng=self.rng, n_nums=n_nums, n_chunk=n_chunk, num_type=num_type, 
                                                 num_min=num_min, num_max=num_max, out_file=True, out_path=out_path, 
                                                 out_file_binary=out_binary, out_file_separator=out_separator, 
                                                 threaded=True)
        result_future.add_done_callback(self.acquisition_finished)


rava_subapp_acquisition = {'class':RAVA_SUBAPP_ACQUISITION,
                           'menu_title':'Acquisition',
                           'show_button':True,
                           'use_rng':True
                           }


'''
This example showcases RNG pulse count measurements for different values of the
sampling interval configuration.

This example code is in the public domain.
Author: Gabriel Guerrer
'''

import rng_rava as rava

# Find RAVA device and connect
rng = rava.RAVA_RNG()
dev_sns = rava.find_rava_sns()
if len(dev_sns):
    rng.connect(serial_number=dev_sns[0])
else:
    print('No device found')
    exit()

# Config PWM
rng.snd_pwm_setup(freq_id=rava.D_PWM_FREQ['50_KHZ'], duty=20)

# Get initial sampling interval
si0 = rng.get_rng_setup()

# Vary sampling intervals
sampling_intervals = range(1, 20+1)

for si in sampling_intervals:
    rng.snd_rng_setup(sampling_interval_us=si)

    # Measure pulse counts
    pcs_a, pcs_b = rng.get_rng_pulse_counts(n_counts=5000, output_type='array')

    # Calculate mean values
    pcs_mean_a = pcs_a.mean()
    pcs_mean_b = pcs_b.mean()

    # Inform mean values
    print('\nsi={} us; pc_a={:.2f}, pc_b={:.2f}'.format(si, pcs_mean_a, pcs_mean_b))

# Restore initial sampling interval
rng.snd_rng_setup(**si0)

# Close device
rng.close()


'''
This file should be opened in an interactive Python environment. In VSCode, the
users can utilize the F6 shortcut to execute the initial code, establishing a
connection with the device. Next, users can navigate the code and execute each
line by positioning the cursor and pressing Shift + Enter. This approach
facilitates comprehensive testing of the RAVA's functionality.

This example code is in the public domain.
Author: Gabriel Guerrer
'''

import rng_rava as rava

# Create RAVA instance and set logging level to INFO
rng = rava.RAVA_RNG_LED()
rng.log_level('INFO')

# Find RAVA device and connect
rava_sns = rava.find_rava_sns()
if len(rava_sns):
    rng.connect(serial_number=rava_sns[0])
else:
    print('No device found')


def DEVICE():

    rng.connect(serial_number=rava_sns[0])
    rng.connected()
    rng.close()

    rng.get_device_serial_number()
    rng.get_device_temperature()
    rng.get_device_free_ram()

    rng.snd_device_reboot()


def EEPROM():

    rng.snd_eeprom_reset_to_default()

    rng.get_eeprom_device()
    rng.snd_eeprom_device(temp_calib_slope=397, temp_calib_intercept=-280)
    rng.snd_eeprom_device(temp_calib_slope=1, temp_calib_intercept=0)

    rng.get_eeprom_firmware()

    rng.get_eeprom_pwm()
    rng.snd_eeprom_pwm(freq_id=rava.D_PWM_FREQ['30_KHZ'], duty=12)
    rng.snd_eeprom_pwm(freq_id=rava.D_PWM_FREQ['40_KHZ'], duty=12)
    rng.snd_eeprom_pwm(freq_id=rava.D_PWM_FREQ['50_KHZ'], duty=20)
    rng.snd_eeprom_pwm(freq_id=rava.D_PWM_FREQ['60_KHZ'], duty=25)
    rng.snd_eeprom_pwm(freq_id=rava.D_PWM_FREQ['75_KHZ'], duty=25)

    rng.get_eeprom_rng()
    rng.snd_eeprom_rng(sampling_interval_us=5)
    rng.snd_eeprom_rng(sampling_interval_us=10)
    rng.snd_eeprom_rng(sampling_interval_us=15)

    rng.get_eeprom_led()
    rng.snd_eeprom_led(led_attached=False)
    rng.snd_eeprom_led(led_attached=True)

    rng.get_eeprom_lamp()
    rng.snd_eeprom_lamp(exp_dur_max_ms=5*60*1000, exp_z_significant=3.925, exp_mag_smooth_n_trials=20)
    rng.snd_eeprom_lamp(exp_dur_max_ms=1*60*1000, exp_z_significant=3.925, exp_mag_smooth_n_trials=10)


def PWM():

    rng.get_pwm_setup()

    rng.snd_pwm_setup(freq_id=rava.D_PWM_FREQ['30_KHZ'], duty=12)
    rng.snd_pwm_setup(freq_id=rava.D_PWM_FREQ['40_KHZ'], duty=12)
    rng.snd_pwm_setup(freq_id=rava.D_PWM_FREQ['50_KHZ'], duty=20)
    rng.snd_pwm_setup(freq_id=rava.D_PWM_FREQ['60_KHZ'], duty=25)
    rng.snd_pwm_setup(freq_id=rava.D_PWM_FREQ['75_KHZ'], duty=25)


def RNG():

    rng.get_rng_setup()

    rng.snd_rng_setup(sampling_interval_us=5)
    rng.snd_rng_setup(sampling_interval_us=10)
    rng.snd_rng_setup(sampling_interval_us=15)

    rng.snd_rng_timing_debug_d1(on=True)
    rng.snd_rng_timing_debug_d1(on=False)

    rng.get_rng_pulse_counts(n_counts=15, output_type='list')

    rng.get_rng_bits(bit_source_id=rava.D_RNG_BIT_SRC['AB'])
    rng.get_rng_bits(bit_source_id=rava.D_RNG_BIT_SRC['A'])
    rng.get_rng_bits(bit_source_id=rava.D_RNG_BIT_SRC['B'])
    rng.get_rng_bits(bit_source_id=rava.D_RNG_BIT_SRC['AB_XOR'])
    rng.get_rng_bits(bit_source_id=rava.D_RNG_BIT_SRC['AB_RND'])

    rng.get_rng_bytes(n_bytes=15, postproc_id=rava.D_RNG_POSTPROC['NONE'], output_type='array')
    rng.get_rng_bytes(n_bytes=15, postproc_id=rava.D_RNG_POSTPROC['XOR'], output_type='array')
    rng.get_rng_bytes(n_bytes=15, postproc_id=rava.D_RNG_POSTPROC['XOR_DICHTL'], output_type='array')
    rng.get_rng_bytes(n_bytes=15, postproc_id=rava.D_RNG_POSTPROC['VON_NEUMANN'], output_type='array')

    rng.get_rng_int8s(n_ints=15, int_delta=10)
    rng.get_rng_int8s(n_ints=15, int_delta=100)
    rng.get_rng_int16s(n_ints=15, int_delta=1000)
    rng.get_rng_floats(n_floats=15)

    rng.get_rng_byte_stream_status()
    rng.snd_rng_byte_stream_start(n_bytes=10, stream_interval_ms=200, postproc_id=rava.D_RNG_POSTPROC['NONE'])
    rng.get_rng_byte_stream_data(output_type='array')
    rng.snd_rng_byte_stream_stop()


def HEALTH():

    rng.snd_health_startup_run()
    rng.get_health_startup_results()
    rava.print_health_startup_results(*rng.get_health_startup_results())

    rng.get_health_continuous_errors()


def PERIPHERALS_DIGITAL_IO():

    rng.snd_periph_digi_mode(periph_id=1, mode_id=rava.D_PERIPH_MODES['INPUT'])
    rng.snd_periph_digi_mode(periph_id=1, mode_id=rava.D_PERIPH_MODES['OUTPUT'])
    rng.get_periph_digi_state(periph_id=1)
    rng.snd_periph_digi_state(periph_id=1, digi_state=0)
    rng.snd_periph_digi_state(periph_id=1, digi_state=1)
    rng.snd_periph_digi_pulse(periph_id=1, pulse_duration_us=100)

    rng.snd_periph_digi_mode(periph_id=2, mode_id=rava.D_PERIPH_MODES['INPUT'])
    rng.snd_periph_digi_mode(periph_id=2, mode_id=rava.D_PERIPH_MODES['OUTPUT'])
    rng.get_periph_digi_state(periph_id=2)
    rng.snd_periph_digi_state(periph_id=2, digi_state=0)
    rng.snd_periph_digi_state(periph_id=2, digi_state=1)
    rng.snd_periph_digi_pulse(periph_id=2, pulse_duration_us=100)

    rng.snd_periph_digi_mode(periph_id=3, mode_id=rava.D_PERIPH_MODES['INPUT'])
    rng.snd_periph_digi_mode(periph_id=3, mode_id=rava.D_PERIPH_MODES['OUTPUT'])
    rng.get_periph_digi_state(periph_id=3)
    rng.snd_periph_digi_state(periph_id=3, digi_state=0)
    rng.snd_periph_digi_state(periph_id=3, digi_state=1)
    rng.snd_periph_digi_pulse(periph_id=3, pulse_duration_us=100)

    rng.snd_periph_digi_mode(periph_id=4, mode_id=rava.D_PERIPH_MODES['INPUT'])
    rng.snd_periph_digi_mode(periph_id=4, mode_id=rava.D_PERIPH_MODES['OUTPUT'])
    rng.get_periph_digi_state(periph_id=4)
    rng.snd_periph_digi_state(periph_id=4, digi_state=0)
    rng.snd_periph_digi_state(periph_id=4, digi_state=1)
    rng.snd_periph_digi_pulse(periph_id=4, pulse_duration_us=100)

    rng.snd_periph_digi_mode(periph_id=5, mode_id=rava.D_PERIPH_MODES['INPUT'])
    rng.snd_periph_digi_mode(periph_id=5, mode_id=rava.D_PERIPH_MODES['OUTPUT'])
    rng.get_periph_digi_state(periph_id=5)
    rng.snd_periph_digi_state(periph_id=5, digi_state=0)
    rng.snd_periph_digi_state(periph_id=5, digi_state=1)
    rng.snd_periph_digi_pulse(periph_id=5, pulse_duration_us=100)


def PERIPHERALS():

    rng.snd_periph_d1_trigger_input(on=True)
    rng.snd_periph_d1_trigger_input(on=False)

    rng.snd_periph_d1_comparator(neg_to_d5=False)
    rng.snd_periph_d1_comparator(neg_to_d5=True)
    rng.snd_periph_d1_comparator(on=False)

    rng.snd_periph_digi_mode(periph_id=1, mode_id=rava.D_PERIPH_MODES['OUTPUT'])
    rng.snd_periph_d1_delay_us_test(delay_us=1)
    rng.snd_periph_d1_delay_us_test(delay_us=5)
    rng.snd_periph_d1_delay_us_test(delay_us=10)

    # Running with an unconnected D2 may flood the driver with random signaling
    rng.snd_periph_d2_input_capture(on=True)
    rng.snd_periph_d2_input_capture(on=False)
    rng.get_periph_d2_input_capture()

    rng.snd_periph_d3_timer3_trigger_output(interval_ms=1)
    rng.snd_periph_d3_timer3_trigger_output(interval_ms=10)
    rng.snd_periph_d3_timer3_trigger_output(interval_ms=100)
    rng.snd_periph_d3_timer3_trigger_output(on=False)

    rng.snd_periph_d3_timer3_pwm(freq_prescaler=1, top=2**16-1, duty=1000)
    rng.snd_periph_d3_timer3_pwm(freq_prescaler=1, top=2**8-1, duty=10)
    rng.snd_periph_d3_timer3_pwm(freq_prescaler=1, top=2**5-1, duty=10)
    rng.snd_periph_d3_timer3_pwm(on=False)

    rng.snd_periph_d4_pin_change(on=True)
    rng.snd_periph_d4_pin_change(on=False)

    rng.get_periph_d5_adc_read(ref_5v=0, clk_prescaler=6, oversampling_n_bits=0)
    rng.get_periph_d5_adc_read(ref_5v=1, clk_prescaler=6, oversampling_n_bits=0)
    rng.get_periph_d5_adc_read(ref_5v=0, clk_prescaler=6, oversampling_n_bits=6)
    rng.get_periph_d5_adc_read(ref_5v=1, clk_prescaler=6, oversampling_n_bits=6)


def INTERFACES():

    rng.get_interface_ds18bs0()


def LED():

    rng.snd_led_color(color_hue=rava.D_LED_COLOR['RED'], intensity=255)
    rng.snd_led_color(color_hue=rava.D_LED_COLOR['ORANGE'], intensity=255)
    rng.snd_led_color(color_hue=rava.D_LED_COLOR['YELLOW'], intensity=255)
    rng.snd_led_color(color_hue=rava.D_LED_COLOR['GREEN'], intensity=255)
    rng.snd_led_color(color_hue=rava.D_LED_COLOR['BLUE'], intensity=255)
    rng.snd_led_color(color_hue=rava.D_LED_COLOR['PURPLE'], intensity=255)
    rng.snd_led_color(color_hue=rava.D_LED_COLOR['PINK'], intensity=255)

    rng.snd_led_color_fade(color_hue_tgt=rava.D_LED_COLOR['RED'], duration_ms=1000)
    rng.snd_led_color_fade(color_hue_tgt=rava.D_LED_COLOR['ORANGE'], duration_ms=1000)
    rng.snd_led_color_fade(color_hue_tgt=rava.D_LED_COLOR['YELLOW'], duration_ms=1000)
    rng.snd_led_color_fade(color_hue_tgt=rava.D_LED_COLOR['GREEN'], duration_ms=1000)
    rng.snd_led_color_fade(color_hue_tgt=rava.D_LED_COLOR['BLUE'], duration_ms=1000)
    rng.snd_led_color_fade(color_hue_tgt=rava.D_LED_COLOR['PINK'], duration_ms=1000)

    rng.snd_led_color_oscillate(n_cycles=3, duration_ms=3000)
    rng.snd_led_color_oscillate(n_cycles=10, duration_ms=10000)

    rng.snd_led_intensity(intensity=0)
    rng.snd_led_intensity(intensity=63)
    rng.snd_led_intensity(intensity=127)
    rng.snd_led_intensity(intensity=190)
    rng.snd_led_intensity(intensity=255)

    rng.snd_led_intensity_fade(intensity_tgt=0, duration_ms=1000)
    rng.snd_led_intensity_fade(intensity_tgt=255, duration_ms=1000)

    rng.snd_led_fade_stop()

    rng.get_led_status()


def LAMP():

    rng.snd_lamp_mode(on=True)
    rng.snd_lamp_mode(on=False)

    rng.snd_lamp_debug(on=True)
    rng.get_lamp_debug_data()
    rng.snd_lamp_debug(on=False)

    rng.get_lamp_statistics()


'''
This example illustrates the generation of random bytes to binary files using 
the RAVA_ACQUISITION class.

This example code is in the public domain.
Author: Gabriel Guerrer
'''

import rng_rava as rava

# Variables
OUT_FILE = True                         # File output
OUT_PATH = './'                         # Update with the desired output path
N_BYTES = 100000                        # 100KB
N_CHUNK = 10000                         # Update progress every 10KB
POSTPROC = 'NONE'                       # No post-processing
RNG_OUT = 'AB'                          # Generate 2 files, one for each RNG core
THREADED = False                        # Blocking mode

# Find RAVA device and connect
rng = rava.RAVA_RNG()
dev_sns = rava.find_rava_sns()
if len(dev_sns):
    rng.connect(serial_number=dev_sns[0])
else:
    print('No device found')
    exit()

# Create acquisition instance
rng_acq = rava.acq.RAVA_ACQUISITION()

# Register progress callback
rng_acq.cbkreg_progress(lambda progress: print('{:.0f}%'.format(progress)))

# Acquisition
outputs, progress, time_str = rng_acq.get_bytes(rng, N_BYTES, N_CHUNK, POSTPROC, RNG_OUT, OUT_FILE, OUT_PATH, THREADED)

# Close RAVA device
rng.close()